import collections

import pytest
from spacy.tokens import Span, Token

from textacy import load_spacy_lang
from textacy import extract
from textacy.extract import basics


@pytest.fixture(scope="module")
def spacy_doc():
    spacy_lang = load_spacy_lang("en_core_web_sm")
    text = (
        "Two weeks ago, I was in Kuwait participating in an I.M.F. (International Monetary Fund) seminar for Arab educators. "
        "For 30 minutes, we discussed the impact of technology trends on education in the Middle East. "
        "And then an Egyptian education official raised his hand and asked if he could ask me a personal question: \"I heard Donald Trump say we need to close mosques in the United States,\" he said with great sorrow. "
        "\"Is that what we want our kids to learn?\""
    )
    spacy_doc = spacy_lang(text)
    return spacy_doc


class TestWords:

    def test_default(self, spacy_doc):
        result = list(extract.words(spacy_doc))
        assert all(isinstance(tok, Token) for tok in result)
        assert not any(tok.is_space for tok in result)

    def test_filter(self, spacy_doc):
        result = list(
            extract.words(
                spacy_doc, filter_stops=True, filter_punct=True, filter_nums=True
            )
        )
        assert not any(tok.is_stop for tok in result)
        assert not any(tok.is_punct for tok in result)
        assert not any(tok.like_num for tok in result)

    def test_pos(self, spacy_doc):
        result1 = list(extract.words(spacy_doc, include_pos={"NOUN"}))
        result2 = list(extract.words(spacy_doc, include_pos="NOUN"))
        assert all(tok.pos_ == "NOUN" for tok in result1)
        assert all(tok.pos_ == "NOUN" for tok in result2)
        result3 = list(extract.words(spacy_doc, exclude_pos={"NOUN"}))
        result4 = list(extract.words(spacy_doc, exclude_pos="NOUN"))
        assert not any(tok.pos_ == "NOUN" for tok in result3)
        assert not any(tok.pos_ == "NOUN" for tok in result4)

    def test_min_freq(self, spacy_doc):
        counts = collections.Counter()
        counts.update(tok.lower_ for tok in spacy_doc)
        result = list(extract.words(spacy_doc, min_freq=2))
        assert all(counts[tok.lower_] >= 2 for tok in result)


class TestNGrams:

    @pytest.mark.parametrize("n", [1, 2])
    def test_n(self, n, spacy_doc):
        result = list(extract.ngrams(spacy_doc, n))
        assert all(isinstance(span, Span) for span in result)
        assert all(len(span) == n for span in result)

    @pytest.mark.parametrize("ns", [[1, 2], [1, 2, 3]])
    def test_multiple_ns(self, ns, spacy_doc):
        result = list(extract.ngrams(spacy_doc, ns))
        assert all(isinstance(span, Span) for span in result)
        minn = min(ns)
        maxn = max(ns)
        assert all(minn <= len(span) <= maxn for span in result)

    def test_n_less_than_1(self, spacy_doc):
        with pytest.raises(ValueError):
            _ = list(extract.ngrams(spacy_doc, 0))

    def test_filter(self, spacy_doc):
        result = list(
            extract.ngrams(
                spacy_doc, 2, filter_stops=True, filter_punct=True, filter_nums=True
            )
        )
        assert not any(span[0].is_stop or span[-1].is_stop for span in result)
        assert not any(tok.is_punct for span in result for tok in span)
        assert not any(tok.like_num for span in result for tok in span)

    def test_min_freq(self, spacy_doc):
        n = 2
        counts = collections.Counter()
        counts.update(spacy_doc[i : i + n].text.lower() for i in range(len(spacy_doc) - n + 1))
        result = list(extract.ngrams(spacy_doc, 2, min_freq=2))
        assert all(counts[span.text.lower()] >= 2 for span in result)

    def test_pos(self, spacy_doc):
        result1 = list(extract.ngrams(spacy_doc, 2, include_pos={"NOUN"}))
        result2 = list(extract.ngrams(spacy_doc, 2, include_pos="NOUN"))
        assert all(tok.pos_ == "NOUN" for span in result1 for tok in span)
        assert all(tok.pos_ == "NOUN" for span in result2 for tok in span)
        result3 = list(extract.ngrams(spacy_doc, 2, exclude_pos={"NOUN"}))
        result4 = list(extract.ngrams(spacy_doc, 2, exclude_pos="NOUN"))
        assert not any(tok.pos_ == "NOUN" for span in result3 for tok in span)
        assert not any(tok.pos_ == "NOUN" for span in result4 for tok in span)


class TestEntities:

    def test_default(self, spacy_doc):
        result = list(extract.entities(spacy_doc, drop_determiners=False))
        assert all(isinstance(span, Span) for span in result)
        assert all(span.label_ for span in result)
        assert all(span[0].ent_type for span in result)

    def test_include_types(self, spacy_doc):
        ent_types = ["PERSON", "GPE"]
        for include_types in ent_types:
            result = extract.entities(spacy_doc, include_types=include_types)
            assert all(span.label_ == include_types for span in result)
        ent_types = [{"PERSON", "GPE"}, ("DATE", "ORG"), ["LOC"]]
        for include_types in ent_types:
            result = extract.entities(spacy_doc, include_types=include_types)
            assert all(span.label_ in include_types for span in result)
        # special numeric cases!
        ent_types = ["NUMERIC", ("NUMERIC",), {"PERSON", "NUMERIC"}]
        for include_types in ent_types:
            include_types_parsed = basics._parse_ent_types(include_types, "include")
            result = extract.entities(spacy_doc, include_types=include_types)
            assert all(span.label_ in include_types_parsed for span in result)

    def test_exclude_types(self, spacy_doc):
        ent_types = ["PERSON", "GPE"]
        for exclude_types in ent_types:
            result = extract.entities(spacy_doc, exclude_types=exclude_types)
            assert all(span.label_ != exclude_types for span in result)
        ent_types = [{"PERSON", "GPE"}, ("DATE", "ORG"), ["LOC"]]
        for exclude_types in ent_types:
            result = extract.entities(spacy_doc, exclude_types=exclude_types)
            assert all(span.label_ not in exclude_types for span in result)
        # special numeric cases!
        ent_types = ["NUMERIC", ("NUMERIC",), {"PERSON", "NUMERIC"}]
        for exclude_types in ent_types:
            exclude_types_parsed = basics._parse_ent_types(exclude_types, "exclude")
            result = extract.entities(spacy_doc, exclude_types=exclude_types)
            assert all(span.label_ not in exclude_types_parsed for span in result)

    def test_parse_ent_types_bad_type(self):
        for bad_type in [1, 3.1415, True, b"PERSON"]:
            with pytest.raises(TypeError):
                _ = basics._parse_ent_types(bad_type, "include")

    def test_min_freq(self, spacy_doc):
        result = list(extract.entities(spacy_doc, min_freq=2))
        assert len(result) == 0

    def test_determiner(self, spacy_doc):
        result = list(extract.entities(spacy_doc, drop_determiners=False))
        assert all(isinstance(span, Span) for span in result)
        assert any(span[0].pos_ == "DET" for span in result)
        assert all(span.label_ for span in result)

    def test_drop_determiners(self, spacy_doc):
        result = list(extract.entities(spacy_doc, drop_determiners=True))
        assert not any(span[0].pos_ == "DET" for span in result)
        assert all(span.label_ for span in result)


class TestNounChunks:

    def test_default(self, spacy_doc):
        result = list(extract.noun_chunks(spacy_doc))
        assert all(isinstance(span, Span) for span in result)

    def test_determiner(self, spacy_doc):
        result = list(extract.noun_chunks(spacy_doc, drop_determiners=False))
        assert all(isinstance(span, Span) for span in result)
        assert any(span[0].pos_ == "DET" for span in result)

    def test_min_freq(self, spacy_doc):
        text = spacy_doc.text.lower()
        result = list(extract.noun_chunks(spacy_doc, drop_determiners=True, min_freq=2))
        assert all(text.count(span.text.lower()) >= 2 for span in result)


class TestTerms:

    def test_default(self, spacy_doc):
        with pytest.raises(ValueError):
            _ = list(extract.terms(spacy_doc))

    def test_simple_args(self, spacy_doc):
        results = list(extract.terms(spacy_doc, ngs=2, ents=True, ncs=True))
        assert results
        assert all(isinstance(result, Span) for result in results)

    def test_callable_args(self, spacy_doc):
        results = list(
            extract.terms(
                spacy_doc,
                ngs=lambda doc: extract.ngrams(doc, n=2),
                ents=extract.entities,
                ncs=extract.noun_chunks,
            )
        )
        assert results
        assert all(isinstance(result, Span) for result in results)

    @pytest.mark.parametrize("dedupe", [True, False])
    def test_dedupe(self, dedupe, spacy_doc):
        results = list(extract.terms(spacy_doc, ngs=2, ents=True, ncs=True, dedupe=dedupe))
        assert results
        if dedupe is True:
            assert (
                len(results) ==
                len(set((result.start, result.end) for result in results))
            )
        else:
            assert (
                len(results) >
                len(set((result.start, result.end) for result in results))
            )



import re

import pytest

from textacy import load_spacy_lang
from textacy import extract


@pytest.fixture(scope="module")
def en_nlp():
    return load_spacy_lang("en_core_web_sm")


@pytest.fixture(scope="module")
def es_nlp():
    return load_spacy_lang("es_core_news_sm")


@pytest.fixture(scope="module")
def sss_doc(en_nlp):
    text = (
        "In general, Burton DeWilde loves animals, but he does not love *all* animals. "
        "Burton loves his cats Rico and Isaac; Burton loved his cat Lucy. "
        "But Burton DeWilde definitely does not love snakes, spiders, or moths. "
        "Now you know that Burton loves animals and cats in particular."
    )
    return en_nlp(text)


@pytest.mark.parametrize(
    "text, svos_exp",
    [
        (
            "Burton loves cats.",
            [(["Burton"], ["loves"], ["cats"])],
        ),
        (
            "One of my cats was eating food.",
            [(["One"], ["was", "eating"], ["food"])],
        ),
        (
            "The first dog to arrive wins the treat.",
            [(["dog"], ["wins"], ["treat"])],
        ),
        (
            "Burton was loved by cats.",
            [(["Burton"], ["was", "loved"], ["cats"])],
        ),
        (
            "Food was eaten by my cat.",
            [(["Food"], ["was", "eaten"], ["cat"])],
        ),
        (
            "The treat was won by the first dog to arrive.",
            [(["treat"], ["was", "won"], ["dog"])],
        ),
        (
            "He and I love house cats and big dogs.",
            [(["He", "I"], ["love"], ["house", "cats", "dogs"])],
        ),
        (
            "We do love and did hate small dogs.",
            [(["We"], ["do", "love"], ["dogs"]), (["We"], ["did", "hate"], ["dogs"])],
        ),
        (
            "Rico eats food and plays fetch.",
            [(["Rico"], ["eats"], ["food"]), (["Rico"], ["plays"], ["fetch"])]
        ),
        (
            "What the cat did makes sense.",
            [
                (["cat"], ["did"], ["What"]),
                (["What", "the", "cat", "did"], ["makes"], ["sense"]),
            ],
        ),
        (
            "What the cat wanted was eaten by the dog.",
            [
                (["cat"], ["wanted"], ["What"]),
                (["What", "the", "cat", "wanted"], ["was", "eaten"], ["dog"]),
            ],
        ),
        (
            "Burton and Nick do not expect to adopt another cat.",
            [
                (
                    ["Burton", "Nick"],
                    ["do", "not", "expect"],
                    ["to", "adopt", "another", "cat"],
                )
            ],
        ),
        (
            "She and her friend did find, sell, and throw sea shells by the sea shore.",
            [
                (["She", "friend"], ["did", "find"], ["sea", "shells"]),
                (["She", "friend"], ["sell"], ["sea", "shells"]),
                (["She", "friend"], ["throw"], ["sea", "shells"]),
            ],
        ),
    ]
)
def test_subject_verb_object_triples(text, svos_exp, en_nlp):
    doc = en_nlp(text)
    svos = list(extract.subject_verb_object_triples(doc))
    assert all(
        hasattr(svo, attr)
        for svo in svos for attr in ["subject", "verb", "object"]
    )
    svos_obs = [
        (
            [tok.text for tok in subject],
            [tok.text for tok in verb],
            [tok.text for tok in object]
        )
        for subject, verb, object in svos
    ]
    assert svos_obs == svos_exp


@pytest.mark.parametrize(
    "entity, cue, fragment_len_range, exp",
    [
        (
            "Burton",
            "love",
            None,
            [
                (["Burton"], ["loves"], ["his", "cats", "Rico", "and", "Isaac"]),
                (["Burton"], ["loved"], ["his", "cat", "Lucy"]),
                (["Burton"], ["loves"], ["animals", "and", "cats"])
            ],
        ),
        (
            re.compile("Burton"),
            "love",
            None,
            [
                (["Burton"], ["loves"], ["his", "cats", "Rico", "and", "Isaac"]),
                (["Burton"], ["loved"], ["his", "cat", "Lucy"]),
                (["Burton"], ["loves"], ["animals", "and", "cats"])
            ],
        ),
        (
            "Burton( DeWilde)?",
            "love",
            None,
            [
                (["Burton", "DeWilde"], ["loves"], ["animals"]),
                (["Burton"], ["loves"], ["his", "cats", "Rico", "and", "Isaac"]),
                (["Burton"], ["loved"], ["his", "cat", "Lucy"]),
                (
                    ["Burton", "DeWilde"],
                    ["does", "not", "love"],
                    ["snakes", ",", "spiders", ",", "or", "moths"]
                ),
                (["Burton"], ["loves"], ["animals", "and", "cats"])
            ],
        ),
        (
            "Burton",
            "love",
            (None, 4),
            [
                (["Burton"], ["loved"], ["his", "cat", "Lucy"]),
                (["Burton"], ["loves"], ["animals", "and", "cats"])
            ],
        ),
        (
            "Burton",
            "love",
            (4, 6),
            [(["Burton"], ["loves"], ["his", "cats", "Rico", "and", "Isaac"])]
        ),
        ("Burton", "hate", None, []),
    ],
)
def test_semistructured_statements(sss_doc, entity, cue, fragment_len_range, exp):
    obs = list(
        extract.semistructured_statements(
            sss_doc, entity=entity, cue=cue, fragment_len_range=fragment_len_range
        )
    )
    assert all(
        hasattr(sss, attr)
        for sss in obs for attr in ["entity", "cue", "fragment"]
    )
    obs_text = [
        (
            [tok.text for tok in e],
            [tok.text for tok in c],
            [tok.text for tok in f]
        )
        for e, c, f in obs
    ]
    assert obs_text == exp


@pytest.mark.parametrize(
    "text, exp",
    [
        (
            "Burton said, \"I love cats!\"",
            [(["Burton"], ["said"], "\"I love cats!\"")],
        ),
        (
            "\"We love cats!\" reply Burton and Nick.",
            [(["Burton", "Nick"], ["reply"], "\"We love cats!\"")],
        ),
        (
            "Burton explained from a podium. \"I love cats,\" he said.",
            [(["he"], ["said"], "\"I love cats,\"")],
        ),
        (
            "\"I love cats!\" insists Burton. \"I absolutely do.\"",
            [
                (["Burton"], ["insists"], "\"I love cats!\""),
                (["Burton"], ["insists"], "\"I absolutely do.\"")
            ],
        ),
        (
            "\"Some people say otherwise,\" he conceded.",
            [(["he"], ["conceded"], "\"Some people say otherwise,\"")],
        ),
        (
            "Burton claims that his favorite book is \"One Hundred Years of Solitude\".",
            [],
        ),
        (
            "Burton thinks that cats are \"cuties\".",
            [],
        ),
    ]
)
def test_direct_quotations(en_nlp, text, exp):
    obs = list(extract.direct_quotations(en_nlp(text)))
    assert all(
        hasattr(dq, attr)
        for dq in obs for attr in ["speaker", "cue", "content"]
    )
    obs_text = [
        (
            [tok.text for tok in speaker],
            [tok.text for tok in cue],
            content.text
        )
        for speaker, cue, content in obs
    ]
    assert obs_text == exp


@pytest.mark.parametrize(
    "text, exp",
    [
        (
            "\"Me encantan los gatos\", dijo Burtón.",
            [(["Burtón"], ["dijo"], "\"Me encantan los gatos\"")],
        ),
        (
            "«Me encantan los gatos», afirmó Burtón.",
            [(["Burtón"], ["afirmó"], "«Me encantan los gatos»")],
        ),
    ]
)
def test_direct_quotations_spanish(es_nlp, text, exp):
    obs = extract.direct_quotations(es_nlp(text))
    obs_text = [
        (
            [tok.text for tok in speaker],
            [tok.text for tok in cue],
            content.text
        )
        for speaker, cue, content in obs
    ]
    assert obs_text == exp



from cytoolz import itertoolz
import pytest

import textacy
from textacy.representations import network


@pytest.fixture(scope="module")
def docs():
    lang = textacy.load_spacy_lang("en_core_web_sm")
    texts = [
        "Mary had a little lamb. Its fleece was white as snow.",
        "Everywhere that Mary went the lamb was sure to go.",
        # "It followed her to school one day, which was against the rule.",
        # "It made the children laugh and play to see a lamb at school.",
        # "And so the teacher turned it out, but still it lingered near.",
        # "It waited patiently about until Mary did appear.",
        # "Why does the lamb love Mary so? The eager children cry.",
        # "Mary loves the lamb, you know, the teacher did reply.",
    ]
    return [textacy.make_spacy_doc(text, lang=lang) for text in texts]


class TestCooccurrenceNetwork:

    def test_data_sequence_str(self, docs):
        data = [tok.text for tok in docs[0]]
        graph = network.build_cooccurrence_network(data)
        assert len(graph) == len(set(data))
        assert all("weight" in data for _, _, data in graph.edges(data=True))

    def test_data_sequence_sequence_str(self, docs):
        data = [[tok.lower_ for tok in doc] for doc in docs]
        graph = network.build_cooccurrence_network(data)
        assert len(graph) == len(set(itertoolz.concat(data)))
        assert all("weight" in data for _, _, data in graph.edges(data=True))

    def test_window_size(self, docs):
        data = [tok.text for tok in docs[0]]
        graph_ws2 = network.build_cooccurrence_network(data, window_size=2)
        graph_ws3 = network.build_cooccurrence_network(data, window_size=3)
        assert len(graph_ws2) == len(graph_ws3) == len(set(data))
        assert(
            sum(data["weight"] for _, _, data in graph_ws2.edges(data=True)) <
            sum(data["weight"] for _, _, data in graph_ws3.edges(data=True))
        )

    @pytest.mark.parametrize("edge_weighting", ["count", "binary"])
    def test_edge_weighting(self, edge_weighting, docs):
        data = [tok.text for tok in docs[0]]
        graph = network.build_cooccurrence_network(data, edge_weighting=edge_weighting)
        assert len(graph) == len(set(data))
        if edge_weighting == "count":
            assert(
                all(
                    isinstance(data["weight"], int) and data["weight"] >= 1
                    for _, _, data in graph.edges(data=True)
                )
            )
        elif edge_weighting == "binary":
            assert(all(data["weight"] == 1 for _, _, data in graph.edges(data=True)))


class TestSimilarityNetwork:

    def test_data_sequence_str(self, docs):
        data = [tok.text for tok in docs[0]]
        graph = network.build_similarity_network(data, "levenshtein")
        assert len(graph) == len(set(data))
        assert all("weight" in data for _, _, data in graph.edges(data=True))

    def test_data_sequence_sequence_str(self, docs):
        data = [tuple(tok.lower_ for tok in doc) for doc in docs]
        graph = network.build_similarity_network(data, "jaccard")
        assert len(graph) == len(set(data))
        assert all("weight" in data for _, _, data in graph.edges(data=True))

    @pytest.mark.parametrize("edge_weighting", ["hamming", "levenshtein", "jaro"])
    def test_edit_similarity_metrics(self, edge_weighting, docs):
        data = [tok.text for tok in docs[0]]
        graph = network.build_similarity_network(data, edge_weighting)
        assert len(graph) == len(set(data))
        assert(
            all(
                isinstance(data["weight"], float) and 0.0 <= data["weight"] <= 1.0
                for _, _, data in graph.edges(data=True)
            )
        )

    @pytest.mark.parametrize("edge_weighting", ["jaccard", "cosine", "bag"])
    def test_tok_similarity_metrics(self, edge_weighting, docs):
        data = [tuple(tok.lower_ for tok in doc) for doc in docs]
        graph = network.build_similarity_network(data, edge_weighting)
        assert len(graph) == len(set(data))
        assert(
            all(
                isinstance(data["weight"], float) and 0.0 <= data["weight"] <= 1.0
                for _, _, data in graph.edges(data=True)
            )
        )



from contextlib import ExitStack as does_not_raise
# TODO: when only supporting PY3.7+, use this instead
# from contextlib import nullcontext as does_not_raise

import pyphen
import pytest

import textacy
import textacy.text_stats


@pytest.fixture(scope="module")
def ts_en():
    text = (
        "Many years later, as he faced the firing squad, Colonel Aureliano Buendía was "
        "to remember that distant afternoon when his father took him to discover ice. "
        "At that time Macondo was a village of twenty adobe houses, built on the bank "
        "of a river of clear water that ran along a bed of polished stones, which were "
        "white and enormous, like prehistoric eggs. The world was so recent that many "
        "things lacked names, and in order to indicate them it was necessary to point."
    )
    return textacy.text_stats.TextStats(
        textacy.make_spacy_doc(text, lang="en_core_web_sm")
    )


@pytest.fixture(scope="module")
def ts_es():
    text = (
        "Muchos años después, frente al pelotón de fusilamiento, el coronel Aureliano "
        "Buendía había de recordar aquella tarde remota en que su padre lo llevó a "
        "conocer el hielo. Macondo era entonces una aldea de veinte casas de barro y "
        "cañabrava construidas a la orilla de un río de aguas diáfanas que se precipitaban "
        "por un lecho de piedras pulidas, blancas y enormes como huevos prehistóricos. "
        "El mundo era tan reciente, que muchas cosas carecían de nombre, y para "
        "mencionarlas había que señalarlas con el dedo."
    )
    return textacy.text_stats.TextStats(
        textacy.make_spacy_doc(text, lang="es_core_news_sm")
    )


@pytest.mark.parametrize(
    "lang, attr_name, attr_type, attr_subtype, exp_val",
    [
        ("en", "n_sents", int, None, 3),
        ("en", "n_words", int, None, 84),
        ("en", "n_unique_words", int, None, 66),
        ("en", "n_long_words", int, None, 14),
        ("en", "n_chars_per_word", tuple, int, None),
        ("en", "n_chars", int, None, 372),
        ("en", "n_syllables_per_word", tuple, int, None),
        ("en", "n_syllables", int, None, 111),
        ("en", "n_monosyllable_words", int, None, 66),
        ("en", "n_polysyllable_words", int, None, 8),
        ("en", "entropy", float, None, 5.9071),
        ("es", "n_sents", int, None, 3),
        ("es", "n_words", int, None, 85),
        ("es", "n_unique_words", int, None, 66),
        ("es", "n_long_words", int, None, 24),
        ("es", "n_chars_per_word", tuple, int, None),
        ("es", "n_chars", int, None, 412),
        ("es", "n_syllables_per_word", tuple, int, None),
        ("es", "n_syllables", int, None, 160),
        ("es", "n_monosyllable_words", int, None, 38),
        ("es", "n_polysyllable_words", int, None, 20),
        ("es", "entropy", float, None, 5.8269),
    ]
)
def test_basics_attrs(ts_en, ts_es, lang, attr_name, attr_type, attr_subtype, exp_val):
    # NOTE: this is awkward, and it seems like there should be a better way
    ts = ts_en if lang == "en" else ts_es
    assert hasattr(ts, attr_name)
    obs_val = getattr(ts, attr_name)
    assert isinstance(obs_val, attr_type)
    if attr_subtype is not None:
        assert all(isinstance(ov, attr_subtype) for ov in obs_val)
    if exp_val is not None:
        if attr_type is float:
            assert obs_val == pytest.approx(exp_val, rel=1e-2)
        else:
            assert obs_val == exp_val


@pytest.mark.parametrize(
    "lang, attr_name, exp_val",
    [
        ("en", "automated_readability_index", 13.42857),
        ("en", "automatic_arabic_readability_index", 1261.21),
        ("en", "coleman_liau_index", 9.1818),
        ("en", "flesch_kincaid_grade_level", 10.92285),
        ("en", "flesch_reading_ease", 66.6221),
        ("en", "gulpease_index", 55.4285),
        ("en", "gunning_fog_index", 15.00952),
        ("en", "lix", 44.6666),
        ("en", "mu_legibility_index", 97.236),
        ("en", "perspicuity_index", 96.5100),
        ("en", "smog_index", 12.45797),
        ("en", "wiener_sachtextformel", 5.2418),
        ("es", "automated_readability_index", 15.56631),
        ("es", "automatic_arabic_readability_index", 1393.424),
        ("es", "coleman_liau_index", 11.6549),
        ("es", "flesch_kincaid_grade_level", 17.6717),
        ("es", "flesch_reading_ease", 64.9938),
        ("es", "gulpease_index", 51.1176),
        ("es", "gunning_fog_index", 20.74509),
        ("es", "lix", 56.5686),
        ("es", "mu_legibility_index", 58.2734),
        ("es", "perspicuity_index", 61.2310),
        ("es", "smog_index", 17.87934),
        ("es", "wiener_sachtextformel", 10.6155),
    ]
)
def test_readability_attrs(ts_en, ts_es, lang, attr_name, exp_val):
    # NOTE: this is awkward, and it seems like there should be a better way
    ts = ts_en if lang == "en" else ts_es
    assert hasattr(ts, attr_name)
    assert getattr(ts, attr_name) == pytest.approx(exp_val, rel=0.05)


@pytest.mark.parametrize(
    "lang, context",
    [
        ("en", does_not_raise()),
        ("es", does_not_raise()),
        ("un", pytest.raises(KeyError)),
    ]
)
def test_load_hyphenator(lang, context):
    with context:
        hyphenator = textacy.text_stats.load_hyphenator(lang=lang)
        assert isinstance(hyphenator, pyphen.Pyphen)



"""
:mod:`textacy.corpus`: Class for working with a collection of spaCy ``Doc`` s.
Includes functionality for easily adding, getting, and removing documents;
saving to / loading their data from disk; and tracking basic corpus statistics.
"""
from __future__ import annotations

import collections
import itertools
import logging
import math
from typing import (
    Any,
    Callable,
    Counter,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Union,
)

import numpy as np
import spacy
from cytoolz import itertoolz
from spacy.language import Language
from spacy.tokens import Doc, DocBin

from . import io as tio
from . import errors, extensions, spacier, types, utils


LOGGER = logging.getLogger(__name__)


class Corpus:
    """
    An ordered collection of :class:`spacy.tokens.Doc`, all of the same language
    and sharing the same :class:`spacy.language.Language` processing pipeline
    and vocabulary, with data held *in-memory*.

    Initialize from a ``Language`` name or instance and (optionally) one or a stream
    of texts or (text, metadata) pairs:

    .. code-block:: pycon

        >>> ds = textacy.datasets.CapitolWords()
        >>> records = ds.records(limit=50)
        >>> corpus = textacy.Corpus("en_core_web_sm", data=records)
        >>> print(corpus)
        Corpus(50 docs, 32175 tokens)

    Add or remove documents, with automatic updating of corpus statistics:

    .. code-block:: pycon

        >>> texts = ds.texts(congress=114, limit=25)
        >>> corpus.add(texts)
        >>> corpus.add("If Burton were a member of Congress, here's what he'd say.")
        >>> print(corpus)
        Corpus(76 docs, 55906 tokens)
        >>> corpus.remove(lambda doc: doc._.meta.get("speaker_name") == "Rick Santorum")
        >>> print(corpus)
        Corpus(61 docs, 48567 tokens)

    Get subsets of documents matching your particular use case:

    .. code-block:: pycon

        >>> match_func = lambda doc: doc._.meta.get("speaker_name") == "Bernie Sanders"
        >>> for doc in corpus.get(match_func, limit=3):
        ...     print(doc._.preview)
        Doc(159 tokens: "Mr. Speaker, 480,000 Federal employees are work...")
        Doc(336 tokens: "Mr. Speaker, I thank the gentleman for yielding...")
        Doc(177 tokens: "Mr. Speaker, if we want to understand why in th...")

    Get or remove documents by indexing, too:

    .. code-block:: pycon

        >>> corpus[0]._.preview
        'Doc(159 tokens: "Mr. Speaker, 480,000 Federal employees are work...")'
        >>> [doc._.preview for doc in corpus[:3]]
        ['Doc(159 tokens: "Mr. Speaker, 480,000 Federal employees are work...")',
         'Doc(219 tokens: "Mr. Speaker, a relationship, to work and surviv...")',
         'Doc(336 tokens: "Mr. Speaker, I thank the gentleman for yielding...")']
        >>> del corpus[:5]
        >>> print(corpus)
        Corpus(56 docs, 41573 tokens)

    Compute basic corpus statistics:

    .. code-block:: pycon

        >>> corpus.n_docs, corpus.n_sents, corpus.n_tokens
        (56, 1771, 41573)
        >>> word_counts = corpus.word_counts(as_strings=True)
        >>> sorted(word_counts.items(), key=lambda x: x[1], reverse=True)[:5]
        [('-PRON-', 2553), ('people', 215), ('year', 148), ('Mr.', 139), ('$', 137)]
        >>> word_doc_counts = corpus.word_doc_counts(weighting="freq", as_strings=True)
        >>> sorted(word_doc_counts.items(), key=lambda x: x[1], reverse=True)[:5]
        [('-PRON-', 0.9821428571428571),
         ('Mr.', 0.7678571428571429),
         ('President', 0.5),
         ('people', 0.48214285714285715),
         ('need', 0.44642857142857145)]

    Save corpus data to and load from disk:

    .. code-block:: pycon

        >>> corpus.save("./cw_sample.bin.gz")
        >>> corpus = textacy.Corpus.load("en_core_web_sm", "./cw_sample.bin.gz")
        >>> print(corpus)
        Corpus(56 docs, 41573 tokens)

    Args:
        lang:
            Language with which spaCy processes (or processed) all documents
            added to the corpus, whether as ``data`` now or later.

            Pass the name of a spacy language pipeline (e.g. "en_core_web_sm"),
            or an already-instantiated :class:`spacy.language.Language` object.

            A given / detected language string is then used to instantiate
            a corresponding ``Language`` with all default components enabled.
        data: One or a stream of texts, records, or :class:`spacy.tokens.Doc` s
            to be added to the corpus.

            .. seealso:: :meth:`Corpus.add()`

    Attributes:
        lang
        spacy_lang
        docs
        n_docs
        n_sents
        n_tokens
    """

    lang: str
    spacy_lang: Language
    docs: List[Doc]
    _doc_ids: List[int]
    n_docs: int
    n_sents: int
    n_tokens: int

    def __init__(self, lang: types.LangLike, data: Optional[types.CorpusData] = None):
        self.spacy_lang = spacier.utils.resolve_langlike(lang)
        self.lang = self.spacy_lang.lang
        self.docs = []
        self._doc_ids = []
        self.n_docs = 0
        self.n_sents = 0
        self.n_tokens = 0
        if data is not None:
            self.add(data)

    # dunder

    def __str__(self):
        return f"Corpus({self.n_docs} docs, {self.n_tokens} tokens)"

    def __len__(self):
        return self.n_docs

    def __iter__(self):
        for doc in self.docs:
            yield doc

    def __contains__(self, doc):
        return id(doc) in self._doc_ids

    def __getitem__(self, idx_or_slice):
        return self.docs[idx_or_slice]

    def __delitem__(self, idx_or_slice: int | slice):
        if isinstance(idx_or_slice, int):
            self._remove_one_doc_by_index(idx_or_slice)
        elif isinstance(idx_or_slice, slice):
            start, end, step = idx_or_slice.indices(self.n_docs)
            idxs = range(start, end, step)
            self._remove_many_docs_by_index(idxs)
        else:
            raise TypeError(
                errors.type_invalid_msg(
                    "idx_or_slice", type(idx_or_slice), Union[int, slice]
                )
            )

    # add documents

    def add(self, data: types.CorpusData, batch_size: int = 1000, n_process: int = 1):
        """
        Add one or a stream of texts, records, or :class:`spacy.tokens.Doc` s
        to the corpus, ensuring that all processing is or has already been done
        by the :attr:`Corpus.spacy_lang` pipeline.

        Args:
            data
            batch_size: Number of texts to buffer when processing with spaCy.
            n_process: Number of parallel processors to run when processing.
                If -1, this is set to ``multiprocessing.cpu_count()``.

                .. note:: This feature is only applies when ``data`` is a sequence
                   of texts or records.

        See Also:
            * :meth:`Corpus.add_text()`
            * :meth:`Corpus.add_texts()`
            * :meth:`Corpus.add_record()`
            * :meth:`Corpus.add_records()`
            * :meth:`Corpus.add_doc()`
            * :meth:`Corpus.add_docs()`
        """
        if isinstance(data, str):
            self.add_text(data)
        elif isinstance(data, Doc):
            self.add_doc(data)
        elif utils.is_record(data):
            self.add_record(data)
        elif isinstance(data, collections.abc.Iterable):
            first, data = itertoolz.peek(data)
            if isinstance(first, str):
                self.add_texts(data, batch_size=batch_size, n_process=n_process)
            elif isinstance(first, Doc):
                self.add_docs(data)
            elif utils.is_record(first):
                self.add_records(data, batch_size=batch_size, n_process=n_process)
            else:
                raise TypeError(
                    errors.type_invalid_msg("data", type(data), types.CorpusData)
                )
        else:
            raise TypeError(
                errors.type_invalid_msg("data", type(data), types.CorpusData)
            )

    def add_text(self, text: str) -> None:
        """
        Add one text to the corpus, processing it into a :class:`spacy.tokens.Doc`
        using the :attr:`Corpus.spacy_lang` pipeline.

        Args:
            text (str)
        """
        self._add_valid_doc(self.spacy_lang(text))

    def add_texts(
        self, texts: Iterable[str], batch_size: int = 1000, n_process: int = 1,
    ) -> None:
        """
        Add a stream of texts to the corpus, efficiently processing them into
        :class:`spacy.tokens.Doc` s using the :attr:`Corpus.spacy_lang` pipeline.

        Args:
            texts: Sequence of texts to process and add to corpus.
            batch_size: Number of texts to buffer when processing with spaCy.
            n_process: Number of parallel processors to run when processing.
                If -1, this is set to ``multiprocessing.cpu_count()``.

                .. note:: This feature is only available in spaCy 2.2.2+.
        """
        if spacy.__version__ >= "2.2.2":
            for doc in self.spacy_lang.pipe(
                texts, as_tuples=False, batch_size=batch_size, n_process=n_process,
            ):
                self._add_valid_doc(doc)
        else:
            if n_process != 1:
                LOGGER.warning("`n_process` is not available with spacy < 2.2.2")
            for doc in self.spacy_lang.pipe(
                texts, as_tuples=False, batch_size=batch_size,
            ):
                self._add_valid_doc(doc)

    def add_record(self, record: types.Record) -> None:
        """
        Add one record to the corpus, processing it into a :class:`spacy.tokens.Doc`
        using the :attr:`Corpus.spacy_lang` pipeline.

        Args:
            record
        """
        doc = self.spacy_lang(record[0])
        doc._.meta = record[1]
        self._add_valid_doc(doc)

    def add_records(
        self,
        records: Iterable[types.Record],
        batch_size: int = 1000,
        n_process: int = 1,
    ) -> None:
        """
        Add a stream of records to the corpus, efficiently processing them into
        :class:`spacy.tokens.Doc` s using the :attr:`Corpus.spacy_lang` pipeline.

        Args:
            records: Sequence of records to process and add to corpus.
            batch_size: Number of texts to buffer when processing with spaCy.
            n_process: Number of parallel processors to run when processing.
                If -1, this is set to ``multiprocessing.cpu_count()``.

                .. note:: This feature is only available in spaCy 2.2.2+.
        """
        if spacy.__version__ >= "2.2.2":
            for doc, meta in self.spacy_lang.pipe(
                records, as_tuples=True, batch_size=batch_size, n_process=n_process,
            ):
                doc._.meta = meta
                self._add_valid_doc(doc)
        else:
            if n_process != 1:
                LOGGER.warning("`n_process` is not available with spacy < 2.2.2")
            for doc, meta in self.spacy_lang.pipe(
                records, as_tuples=True, batch_size=batch_size,
            ):
                doc._.meta = meta
                self._add_valid_doc(doc)

    def add_doc(self, doc: Doc) -> None:
        """
        Add one :class:`spacy.tokens.Doc` to the corpus, provided it was processed
        using the :attr:`Corpus.spacy_lang` pipeline.

        Args:
            doc
        """
        if not isinstance(doc, Doc):
            raise TypeError(errors.type_invalid_msg("doc", type(doc), Doc))
        if doc.vocab is not self.spacy_lang.vocab:
            raise ValueError(
                f"doc.vocab ({doc.vocab}) must be the same as "
                f"corpus.vocab ({self.spacy_lang.vocab})"
            )
        self._add_valid_doc(doc)

    def add_docs(self, docs: Iterable[Doc]) -> None:
        """
        Add a stream of :class:`spacy.tokens.Doc` s to the corpus, provided
        they were processed using the :attr:`Corpus.spacy_lang` pipeline.

        Args:
            docs
        """
        for doc in docs:
            self.add_doc(doc)

    def _add_valid_doc(self, doc: Doc) -> None:
        self.docs.append(doc)
        self._doc_ids.append(id(doc))
        self.n_docs += 1
        self.n_tokens += len(doc)
        if doc.has_annotation("SENT_START"):
            self.n_sents += itertoolz.count(doc.sents)

    # get documents

    def get(
        self, match_func: Callable[[Doc], bool], limit: Optional[int] = None,
    ) -> Iterator[Doc]:
        """
        Get all (or N <= ``limit``) docs in :class:`Corpus` for which
        ``match_func(doc)`` is True.

        Args:
            match_func: Function that takes a :class:`spacy.tokens.Doc`
                as input and returns a boolean value. For example::

                    Corpus.get(lambda x: len(x) >= 100)

                gets all docs with at least 100 tokens. And::

                    Corpus.get(lambda doc: doc._.meta["author"] == "Burton DeWilde")

                gets all docs whose author was given as 'Burton DeWilde'.
            limit: Maximum number of matched docs to return.

        Yields:
            :class:`spacy.tokens.Doc`: Next document passing ``match_func``.

        .. tip:: To get doc(s) by index, treat :class:`Corpus` as a list and use
           Python's usual indexing and slicing: ``Corpus[0]`` gets the first
           document in the corpus; ``Corpus[:5]`` gets the first 5; etc.
        """
        matched_docs = (doc for doc in self if match_func(doc) is True)
        for doc in itertools.islice(matched_docs, limit):
            yield doc

    # remove documents

    def remove(
        self, match_func: Callable[[Doc], bool], limit: Optional[int] = None,
    ) -> None:
        """
        Remove all (or N <= ``limit``) docs in :class:`Corpus` for which
        ``match_func(doc)`` is True. Corpus doc/sent/token counts are adjusted
        accordingly.

        Args:
            match_func: Function that takes a :class:`spacy.tokens.Doc`
                and returns a boolean value. For example::

                    Corpus.remove(lambda x: len(x) >= 100)

                removes docs with at least 100 tokens. And::

                    Corpus.remove(lambda doc: doc._.meta["author"] == "Burton DeWilde")

                removes docs whose author was given as "Burton DeWilde".
            limit: Maximum number of matched docs to remove.

        .. tip:: To remove doc(s) by index, treat :class:`Corpus` as a list and use
           Python's usual indexing and slicing: ``del Corpus[0]`` removes the
           first document in the corpus; ``del Corpus[:5]`` removes the first
           5; etc.
        """
        matched_docs = (doc for doc in self if match_func(doc) is True)
        self._remove_many_docs_by_index(
            self._doc_ids.index(id(doc)) for doc in itertools.islice(matched_docs, limit)
        )

    def _remove_many_docs_by_index(self, idxs: Iterable[int]) -> None:
        for idx in sorted(idxs, reverse=True):
            self._remove_one_doc_by_index(idx)

    def _remove_one_doc_by_index(self, idx: int) -> None:
        doc = self.docs[idx]
        self.n_docs -= 1
        self.n_tokens -= len(doc)
        if doc.has_annotation("SENT_START"):
            self.n_sents -= itertoolz.count(doc.sents)
        del self.docs[idx]
        del self._doc_ids[idx]

    # useful properties

    @property
    def vectors(self) -> np.ndarray:
        """Constituent docs' word vectors stacked in a 2d array."""
        return np.vstack([doc.vector for doc in self])

    @property
    def vector_norms(self) -> np.ndarray:
        """Constituent docs' L2-normalized word vectors stacked in a 2d array."""
        return np.vstack([doc.vector_norm for doc in self])

    # useful methods

    def word_counts(
        self,
        *,
        by: str = "lemma",  # Literal["lemma", "lower", "norm", "orth"]
        weighting: str = "count",  # Literal["count", "freq"]
        **kwargs,
    ) -> Dict[int, int | float] | Dict[str, int | float]:
        """
        Map the set of unique words in :class:`Corpus` to their counts as
        absolute, relative, or binary frequencies of occurence, similar to
        :meth:`Doc._.to_bag_of_words() <textacy.extensions.to_bag_of_words>`
        but aggregated over all docs.

        Args:
            by: Attribute by which spaCy ``Token`` s are grouped before counting,
                as given by ``getattr(token, by)``.
                If "lemma", tokens are grouped by their base form w/o inflections;
                if "lower", by the lowercase form of the token text;
                if "norm", by the normalized form of the token text;
                if "orth", by the token text exactly as it appears in documents.
                To output keys as strings, append an underscore to any of these options;
                for example, "lemma_" groups tokens by their lemmas as strings.
            weighting: Type of weighting to assign to unique words given by ``by``.
                If "count", weights are the absolute number of occurrences (i.e. counts);
                if "freq", weights are counts normalized by the total token count,
                giving their relative frequency of occurrence.
            **kwargs: Passed directly on to :func:`textacy.extract.words()`
                - filter_stops: If True, stop words are removed before counting.
                - filter_punct: If True, punctuation tokens are removed before counting.
                - filter_nums: If True, number-like tokens are removed before counting.

        Returns:
            Mapping of a unique word id or string (depending on the value of ``by``)
            to its absolute, relative, or binary frequency of occurrence
            (depending on the value of ``weighting``).

        See Also:
            :func:`textacy.representations.matrix_utils.get_term_freqs()`
        """
        word_counts_: Union[Counter[Any], Dict[Any, Union[int, float]]]
        word_counts_ = collections.Counter()
        for doc in self:
            word_counts_.update(
                extensions.to_bag_of_words(doc, by=by, weighting="count", **kwargs)
            )
        if weighting == "count":
            word_counts_ = dict(word_counts_)
        elif weighting == "freq":
            n_tokens = self.n_tokens
            word_counts_ = {
                word: count / n_tokens for word, count in word_counts_.items()
            }
        else:
            raise ValueError(
                errors.value_invalid_msg("weighting", weighting, {"count", "freq"})
            )
        return word_counts_

    def word_doc_counts(
        self,
        *,
        by: str = "lemma",  # Literal["lemma", "lower", "norm", "orth"]
        weighting: str = "count",  # Literal["count", "freq", "idf"]
        smooth_idf: bool = True,
        **kwargs
    ) -> Dict[int, int | float] | Dict[str, int | float]:
        """
        Map the set of unique words in :class:`Corpus` to their *document* counts
        as absolute, relative, or inverse frequencies of occurence.

        Args:
            by: Attribute by which spaCy ``Token`` s are grouped before counting,
                as given by ``getattr(token, by)``.
                If "lemma", tokens are grouped by their base form w/o inflections;
                if "lower", by the lowercase form of the token text;
                if "norm", by the normalized form of the token text;
                if "orth", by the token text exactly as it appears in documents.
                To output keys as strings, append an underscore to any of these options;
                for example, "lemma_" groups tokens by their lemmas as strings.
            weighting: Type of weighting to assign to unique words given by ``by``.
                If "count", weights are the absolute number of occurrences (i.e. counts);
                if "freq", weights are counts normalized by the total token count,
                giving their relative frequency of occurrence;
                if "idf", weights are the log of the inverse relative frequencies, i.e.
                ``log(n_docs / word_doc_count)`` or, if ``smooth_idf`` is True,
                ``log(1 + (n_docs / word_doc_count))``.
            smooth_idf: If True, add 1 to all word doc counts when
                calculating "idf" weighting, equivalent to adding a single
                document to the corpus containing every unique word.

        Returns:
            Mapping of a unique word id or string (depending on the value of ``by``)
            to the number of documents in which it appears,
            weighted as absolute, relative, or inverse frequency of occurrence
            (depending on the value of ``weighting``).

        See Also:
            :func:`textacy.vsm.get_doc_freqs() <textacy.vsm.matrix_utils.get_doc_freqs>`
        """
        word_doc_counts_: Union[Counter[Any], Dict[Any, Union[int, float]]]
        word_doc_counts_ = collections.Counter()
        for doc in self:
            word_doc_counts_.update(
                extensions.to_bag_of_words(doc, by=by, weighting="binary", **kwargs)
            )
        if weighting == "count":
            word_doc_counts_ = dict(word_doc_counts_)
        elif weighting == "freq":
            n_docs = self.n_docs
            word_doc_counts_ = {
                word: count / n_docs for word, count in word_doc_counts_.items()
            }
        elif weighting == "idf":
            n_docs = self.n_docs
            if smooth_idf is True:
                word_doc_counts_ = {
                    word: math.log1p(n_docs / count)
                    for word, count in word_doc_counts_.items()
                }
            else:
                word_doc_counts_ = {
                    word: math.log(n_docs / count)
                    for word, count in word_doc_counts_.items()
                }
        else:
            raise ValueError(
                errors.value_invalid_msg(
                    "weighting", weighting, {"count", "freq", "idf"}
                )
            )
        return word_doc_counts_

    def agg_metadata(
        self,
        name: str,
        agg_func: Callable[[Iterable[Any]], Any],
        default: Optional[Any] = None,
    ) -> Any:
        """
        Aggregate values for a particular metadata field over all documents
        in :class:`Corpus`.

        Args:
            name: Name of metadata field (key) in :class:`Doc._.meta`.
            agg_func: Callable that accepts an iterable of field values
                and outputs a single, aggregated result.
            default: Default field value to use if ``name`` is not found
                in a given document's metadata.

        Returns:
            Aggregated value for metadata field.
        """
        return agg_func(doc._.meta.get(name, default) for doc in self)

    # file io

    def save(
        self,
        filepath: types.PathLike,
        attrs: Optional[str | Iterable[str]] = "auto",
        store_user_data: bool = True,
    ):
        """
        Save :class:`Corpus` to disk as binary data.

        Args:
            filepath: Full path to file on disk where :class:`Corpus` docs data
                will be saved as a binary file.
            attrs: List of token attributes to serialize; if "auto", an appropriate list
                is inferred from annotations found on the first ``Doc``; if None,
                spaCy's default values are used (https://spacy.io/api/docbin#init)
            store_user_data: If True, store user data and values of
                custom extension attributes along with core spaCy attributes.

        See Also:
            - :meth:`Corpus.load()`
            - :func:`textacy.io.write_spacy_docs()`
            - :class:`spacy.tokens.DocBin`
        """
        if attrs == "auto":
            doc = self[0]
            attrs = [spacy.attrs.ORTH, spacy.attrs.SPACY]
            cand_attrs = [
                "TAG", "POS", "ENT_IOB", "ENT_TYPE", "ENT_KB_ID", "LEMMA", "MORPH"
            ]
            for cand_attr in cand_attrs:
                if doc.has_annotation(cand_attr):
                    attrs.append(getattr(spacy.attrs, cand_attr))
            # just to be safe, use special handling for dependency parse annotations
            # to allow for multiple ways in which sentences may get segmented
            # *either* use DEP and HEAD or SENT_START to avoid potential conflicts
            # since HEAD implicitly sets sentence boundaries based on tree structure
            if self[0].has_annotation("DEP"):
                attrs.append(spacy.attrs.HEAD)
                attrs.append(spacy.attrs.DEP)
            elif self[0].has_annotation("SENT_START"):
                attrs.append(spacy.attrs.SENT_START)
        tio.write_spacy_docs(
            self.docs,
            filepath,
            format="binary",
            attrs=attrs,
            store_user_data=store_user_data
        )

    @classmethod
    def load(
        cls,
        lang: types.LangLike,
        filepath: types.PathLike,
    ) -> "Corpus":
        """
        Load previously saved :class:`Corpus` binary data, reproduce the original
        `:class:`spacy.tokens.Doc`s tokens and annotations, and instantiate
        a new :class:`Corpus` from them.

        Args:
            lang
            filepath: Full path to file on disk where :class:`Corpus` data
                was previously saved as a binary file.

        Returns:
            Initialized corpus.

        See Also:
            - :meth:`Corpus.save()`
            - :func:`textacy.io.read_spacy_docs()`
            - :class:`spacy.tokens.DocBin`
        """
        spacy_lang = spacier.utils.resolve_langlike(lang)
        docs = tio.read_spacy_docs(filepath, format="binary", lang=spacy_lang)
        return cls(spacy_lang, data=docs)



"""
Token-based Metrics
-------------------

:mod:`textacy.similarity.edits`: Normalized similarity metrics built on token-based
algorithms that identify and count similar tokens between one sequence and another,
and don't rely on the *ordering* of those tokens.
"""
import collections
import math
from typing import Iterable


def jaccard(seq1: Iterable[str], seq2: Iterable[str]) -> float:
    """
    Measure the similarity between two sequences of strings as sets
    using the Jaccard index.

    Args:
        seq1
        seq2

    Returns:
        Similarity between ``seq1`` and ``seq2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar sequences of strings

    Reference:
        https://en.wikipedia.org/wiki/Jaccard_index
    """
    set1 = set(seq1)
    set2 = set(seq2)
    try:
        return len(set1 & set2) / len(set1 | set2)
    except ZeroDivisionError:
        return 0.0


def sorensen_dice(seq1: Iterable[str], seq2: Iterable[str]) -> float:
    """
    Measure the similarity between two sequences of strings as sets
    using the Sørensen-Dice index, which is similar to the Jaccard index.

    Args:
        seq1
        seq2

    Returns:
        Similarity between ``seq1`` and ``seq2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar sequences

    Reference:
        https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient
    """
    set1 = set(seq1)
    set2 = set(seq2)
    try:
        return 2 * len(set1 & set2) / (len(set1) + len(set2))
    except ZeroDivisionError:
        return 0.0


def tversky(
    seq1: Iterable[str], seq2: Iterable[str], *, alpha: float = 1.0, beta: float = 1.0
) -> float:
    """
    Measure the similarity between two sequences of strings as sets
    using the (symmetric) Tversky index, which is a generalization of
    Jaccard (``alpha=0.5, beta=2.0``) and Sørensen-Dice (``alpha=0.5, beta=1.0``).

    Args:
        seq1
        seq2
        alpha
        beta

    Returns:
        Similarity between ``seq1`` and ``seq2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar sequences

    Reference:
        https://en.wikipedia.org/wiki/Tversky_index
    """
    set1 = set(seq1)
    set2 = set(seq2)
    intersection = len(set1 & set2)
    set1_not_set2 = len(set1 - set2)
    set2_not_set1 = len(set2 - set1)
    a = min(set1_not_set2, set2_not_set1)
    b = max(set1_not_set2, set2_not_set1)
    try:
        return intersection / (intersection + (beta * (alpha * a + (1 - alpha) * b)))
    except ZeroDivisionError:
        return 0.0


def cosine(seq1: Iterable[str], seq2: Iterable[str]) -> float:
    """
    Measure the similarity between two sequences of strings as sets
    using the Otsuka-Ochiai variation of cosine similarity (which is equivalent
    to the usual formulation when values are binary).

    Args:
        seq1
        seq2

    Returns:
        Similarity between ``seq1`` and ``seq2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar sequences

    Reference:
        https://en.wikipedia.org/wiki/Cosine_similarity#Otsuka-Ochiai_coefficient
    """
    set1 = set(seq1)
    set2 = set(seq2)
    try:
        return len(set1 & set2) / math.sqrt(len(set1) * len(set2))
    except ZeroDivisionError:
        return 0.0


def bag(seq1: Iterable[str], seq2: Iterable[str]) -> float:
    """
    Measure the similarity between two sequences of strings (*not* as sets)
    using the "bag distance" measure, which can be considered an approximation
    of edit distance.

    Args:
        seq1
        seq2

    Returns:
        Similarity between ``seq1`` and ``seq2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar sequences

    Reference:
        Bartolini, Ilaria, Paolo Ciaccia, and Marco Patella. "String matching with
        metric trees using an approximate distance." International Symposium on String
        Processing and Information Retrieval. Springer, Berlin, Heidelberg, 2002.
    """
    bag1 = collections.Counter(seq1)
    bag2 = collections.Counter(seq2)
    max_diff = max(sum((bag1 - bag2).values()), sum((bag2 - bag1).values()))
    max_len = max(sum(bag1.values()), sum(bag2.values()))
    try:
        return 1.0 - (max_diff / max_len)
    except ZeroDivisionError:
        return 0.0



import pytest

from textacy import make_spacy_doc
from textacy.augmentation import utils


@pytest.fixture(scope="module")
def spacy_doc():
    text = (
        "Democrats might know that they stand against Trump's policies, but coming up with their own plan is harder than you think. "
        "For a long time, the party's top echelon has been captive to free trade orthodoxy. "
        "Since Bill Clinton, the theory of the case among the Democratic Party's elite has been the more globalization, the better — with mostly a deaf ear turned to the people and places most badly affected. "
        "Worse, their response to globalization's excesses has been: "
        "Here's a new trade deal, much better than the last one."
    )
    return make_spacy_doc(text, lang="en_core_web_sm")


def test_aug_tok():
    aug_tok = utils.AugTok(text="text", ws=" ", pos="pos", is_word=True, syns=["doc"])
    assert isinstance(aug_tok, tuple)
    with pytest.raises(AttributeError):
        aug_tok.foo = "bar"


def test_to_aug_toks(spacy_doc):
    aug_toks = utils.to_aug_toks(spacy_doc)
    assert isinstance(aug_toks, list)
    assert all(isinstance(aug_tok, utils.AugTok) for aug_tok in aug_toks)
    assert len(aug_toks) == len(spacy_doc)
    for obj in ["foo bar bat baz", ["foo", "bar", "bat", "baz"]]:
        with pytest.raises(TypeError):
            _ = utils.to_aug_toks(obj)


@pytest.mark.skipif(
    utils.udhr.index is None,
    reason="UDHR dataset must be downloaded before running this test")
def test_get_char_weights():
    for lang in ("en", "es", "xx"):
        char_weights = utils.get_char_weights(lang)
        assert isinstance(char_weights, list)
        assert all(isinstance(item, tuple) for item in char_weights)
        assert all(isinstance(char, str) for char, _ in char_weights)
        assert all(isinstance(weight, (int, float)) for _, weight in char_weights)



from contextlib import ExitStack as does_not_raise
# TODO: when only supporting PY3.7+, use this instead
# from contextlib import nullcontext as does_not_raise

import pytest

from textacy.text_stats import readability


@pytest.fixture(scope="module")
def basics():
    return dict(
        n_sents=3,
        n_words=84,
        n_unique_words=66,
        n_chars=372,
        n_long_words=14,
        n_syllables=113,
        n_monosyllable_words=63,
        n_polysyllable_words=8,
    )


@pytest.mark.parametrize(
    "n_chars, n_words, n_sents, exp",
    [
        (372, 84, 3, 13.4285),
        (200, 100, 2, 12.99),
        (1, 1, 1, -16.22),
    ]
)
def test_automated_readability_index(n_chars, n_words, n_sents, exp):
    obs = readability.automated_readability_index(n_chars, n_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_chars, n_words, n_sents, exp",
    [
        (372, 84, 3, 1261.2128),
        (200, 100, 2, 720.86),
        (1, 1, 1, 5.95),
    ]
)
def test_automatic_arabic_readability_index(n_chars, n_words, n_sents, exp):
    obs = readability.automatic_arabic_readability_index(n_chars, n_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_chars, n_words, n_sents, exp",
    [
        (372, 84, 3, 9.1818),
        (200, 100, 2, -4.6328),
        (1, 1, 1, -39.5082),
    ]
)
def test_coleman_liau_index(n_chars, n_words, n_sents, exp):
    obs = readability.coleman_liau_index(n_chars, n_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_syllables, n_words, n_sents, exp",
    [
        (113, 84, 3, 11.2038),
        (200, 100, 2, 27.51),
        (1, 1, 1, -3.39999),
    ]
)
def test_flesch_kincaid_grade_level(n_syllables, n_words, n_sents, exp):
    obs = readability.flesch_kincaid_grade_level(n_syllables, n_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_syllables, n_words, n_sents, lang, exp",
    [
        (113, 84, 3, "en", 64.6078),
        (200, 100, 2, "en", -13.1149),
        (1, 1, 1, "en", 121.22),
    ]
)
def test_flesch_reading_ease(n_syllables, n_words, n_sents, lang, exp):
    obs = readability.flesch_reading_ease(n_syllables, n_words, n_sents, lang=lang)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "lang,context",
    [
        (None, does_not_raise()),
        ("en", does_not_raise()),
        ("de", does_not_raise()),
        ("es", does_not_raise()),
        ("fr", does_not_raise()),
        ("it", does_not_raise()),
        ("nl", does_not_raise()),
        ("pt", does_not_raise()),
        ("ru", does_not_raise()),
        ("tr", does_not_raise()),
        ("ja", pytest.raises(ValueError)),
        ("un", pytest.raises(ValueError)),
        ("zh", pytest.raises(ValueError)),
    ],
)
def test_flesch_reading_ease_lang(lang, context):
    with context:
        _ = readability.flesch_reading_ease(100, 50, 2, lang=lang)


@pytest.mark.parametrize(
    "n_chars, n_words, n_sents, exp",
    [
        (372, 84, 3, 55.4285),
        (200, 100, 2, 75.0),
        (1, 1, 1, 379.0),
    ]
)
def test_gulpease_index(n_chars, n_words, n_sents, exp):
    obs = readability.gulpease_index(n_chars, n_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_words, n_polysyllable_words, n_sents, exp",
    [
        (84, 8, 3, 15.0095),
        (50, 10, 2, 18.0),
        (1, 1, 1, 40.40),
    ]
)
def test_gunning_fog_index(n_words, n_polysyllable_words, n_sents, exp):
    obs = readability.gunning_fog_index(n_words, n_polysyllable_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_words, n_long_words, n_sents, exp",
    [
        (84, 14, 3, 44.6666),
        (50, 5, 2, 35.0),
        (1, 1, 1, 101.0),
    ]
)
def test_lix(n_words, n_long_words, n_sents, exp):
    obs = readability.lix(n_words, n_long_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_chars_per_word, exp",
    [
        [(4, 5, 5, 2, 2, 5, 3, 6, 5, 7), 180.3278],
        [(6, 4, 7, 6, 2, 7, 2, 12, 2, 7), 62.1468],
        [(1, 2, 1, 2, 1, 2, 1, 2, 1, 2), 600.0],
    ]
)
def test_mu_legibility_index(n_chars_per_word, exp):
    obs = readability.mu_legibility_index(n_chars_per_word)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_syllables, n_words, n_sents, exp",
    [
        (113, 84, 3, 95.02666),
        (200, 100, 2, 32.2350),
        (1, 1, 1, 143.5350),
    ]
)
def test_perspicuity_index(n_syllables, n_words, n_sents, exp):
    obs = readability.perspicuity_index(n_syllables, n_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_polysyllable_words, n_sents, exp",
    [
        (8, 3, 12.4579),
        (10, 2, 15.9031),
        (1, 1, 8.8418),
    ]
)
def test_smog_index(n_polysyllable_words, n_sents, exp):
    obs = readability.smog_index(n_polysyllable_words, n_sents)
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "n_words, n_polysyllable, n_monosyllable, n_long, n_sents, variant, exp",
    [
        (84, 8, 63, 14, 3, 1, 5.3586),
        (84, 8, 63, 14, 3, 2, 6.1303),
        (84, 8, 63, 14, 3, 3, 7.0415),
        (84, 8, 63, 14, 3, 4, 8.3571),
        (1, 1, 1, 1, 1, 1, 28.3422),
    ]
)
def test_wiener_sachtextformel(
    n_words, n_polysyllable, n_monosyllable, n_long, n_sents, variant, exp,
):
    obs = readability.wiener_sachtextformel(
        n_words, n_polysyllable, n_monosyllable, n_long, n_sents, variant=variant,
    )
    assert obs == pytest.approx(exp, rel=1e-2)


@pytest.mark.parametrize(
    "variant, context",
    [
        (0, pytest.raises(ValueError)),
        (1, does_not_raise()),
        (2, does_not_raise()),
        (3, does_not_raise()),
        (4, does_not_raise()),
        (5, pytest.raises(ValueError)),
    ]
)
def test_wiener_sachtextformel_variant(variant, context):
    with context:
        _ = readability.wiener_sachtextformel(84, 8, 63, 14, 3, variant=variant)



"""
Readability Stats
-----------------

:mod:`textacy.text_stats.readability`: Low-level functions for computing various measures
of text "readability", typically accessed via :class:`textacy.text_stats.TextStats`.
"""
import logging
import statistics
from math import sqrt
from typing import Collection, Optional

from .. import errors


LOGGER = logging.getLogger(__name__)

_FRE_COEFS = {
    "de": {"base": 180.0, "asl": 1.0, "awl": 58.5},
    "en": {"base": 206.835, "asl": 1.015, "awl": 84.6},
    "es": {"base": 206.835, "asl": 1.02, "awl": 60.0},  # 0.6 x 100
    "fr": {"base": 207.0, "asl": 1.015, "awl": 73.6},
    "it": {"base": 217.0, "asl": 1.3, "awl": 60.0},  # 0.6 x 100
    "nl": {"base": 206.835, "asl": 0.93, "awl": 77.0},
    "pt": {"base": 248.835, "asl": 1.015, "awl": 84.6},
    "ru": {"base": 206.835, "asl": 1.3, "awl": 60.1},
    "tr": {"base": 198.825, "asl": 2.610, "awl": 40.175},
}


def automated_readability_index(n_chars: int, n_words: int, n_sents: int) -> float:
    """
    Readability test for English-language texts, particularly for technical writing,
    whose value estimates the U.S. grade level required to understand a text.
    Similar to several other tests (e.g. :func:`flesch_kincaid_grade_level()`),
    but uses characters per word instead of syllables like :func:`coleman_liau_index()`.
    Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Automated_readability_index
    """
    return (4.71 * n_chars / n_words) + (0.5 * n_words / n_sents) - 21.43


def automatic_arabic_readability_index(
    n_chars: int, n_words: int, n_sents: int,
) -> float:
    """
    Readability test for Arabic-language texts based on number of characters and
    average word and sentence lengths. Higher value => more difficult text.

    References:
        Al Tamimi, Abdel Karim, et al. "AARI: automatic arabic readability index."
        Int. Arab J. Inf. Technol. 11.4 (2014): 370-378.
    """
    return (3.28 * n_chars) + (1.43 * n_chars / n_words) + (1.24 * n_words / n_sents)


def coleman_liau_index(n_chars: int, n_words: int, n_sents: int) -> float:
    """
    Readability test whose value estimates the number of years of education
    required to understand a text, similar to :func:`flesch_kincaid_grade_level()`
    and :func:`smog_index()`, but using characters per word instead of syllables.
    Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Coleman%E2%80%93Liau_index
    """
    return (5.879851 * n_chars / n_words) - (29.587280 * n_sents / n_words) - 15.800804


def flesch_kincaid_grade_level(n_syllables: int, n_words: int, n_sents: int) -> float:
    """
    Readability test used widely in education, whose value estimates the U.S.
    grade level / number of years of education required to understand a text.
    Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Flesch%E2%80%93Kincaid_readability_tests#Flesch.E2.80.93Kincaid_grade_level
    """  # noqa: E501
    return (11.8 * n_syllables / n_words) + (0.39 * n_words / n_sents) - 15.59


def flesch_reading_ease(
    n_syllables: int, n_words: int, n_sents: int, *, lang: Optional[str] = None,
) -> float:
    """
    Readability test used as a general-purpose standard in several languages, based on
    a weighted combination of avg. sentence length and avg. word length. Values usually
    fall in the range [0, 100], but may be arbitrarily negative in extreme cases.
    Higher value => easier text.

    Note:
        Coefficients in this formula are language-dependent;
        if ``lang`` is null, the English-language formulation is used.

    References:
        English: https://en.wikipedia.org/wiki/Flesch%E2%80%93Kincaid_readability_tests#Flesch_reading_ease
        German: https://de.wikipedia.org/wiki/Lesbarkeitsindex#Flesch-Reading-Ease
        Spanish: Fernández-Huerta formulation
        French: ?
        Italian: https://it.wikipedia.org/wiki/Formula_di_Flesch
        Dutch: ?
        Portuguese: https://pt.wikipedia.org/wiki/Legibilidade_de_Flesch
        Turkish: Atesman formulation
        Russian: https://ru.wikipedia.org/wiki/%D0%98%D0%BD%D0%B4%D0%B5%D0%BA%D1%81_%D1%83%D0%B4%D0%BE%D0%B1%D0%BE%D1%87%D0%B8%D1%82%D0%B0%D0%B5%D0%BC%D0%BE%D1%81%D1%82%D0%B8
    """  # noqa: E501
    try:
        coefs = _FRE_COEFS[lang or "en"]
    except KeyError:
        raise ValueError(errors.value_invalid_msg("lang", lang, list(_FRE_COEFS.keys())))
    return (
        coefs["base"]
        - (coefs["asl"] * n_words / n_sents)
        - (coefs["awl"] * n_syllables / n_words)
    )


def gulpease_index(n_chars: int, n_words: int, n_sents: int) -> float:
    """
    Readability test for Italian-language texts, whose value is in the range
    [0, 100] similar to :func:`flesch_reading_ease()`. Higher value =>
    easier text.

    References:
        https://it.wikipedia.org/wiki/Indice_Gulpease
    """
    return (300 * n_sents / n_words) - (10 * n_chars / n_words) + 89


def gunning_fog_index(n_words: int, n_polysyllable_words: int, n_sents: int) -> float:
    """
    Readability test whose value estimates the number of years of education
    required to understand a text, similar to :func:`flesch_kincaid_grade_level()`
    and :func:`smog_index()`. Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Gunning_fog_index
    """
    return 0.4 * ((n_words / n_sents) + (100 * n_polysyllable_words / n_words))


def lix(n_words: int, n_long_words: int, n_sents: int) -> float:
    """
    Readability test commonly used in Sweden on both English- and non-English-language
    texts, whose value estimates the difficulty of reading a foreign text.
    Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Lix_(readability_test)
    """
    return (n_words / n_sents) + (100 * n_long_words / n_words)


def mu_legibility_index(n_chars_per_word: Collection[int]) -> float:
    """
    Readability test for Spanish-language texts based on number of words and
    the mean and variance of their lengths in characters, whose value is in the range
    [0, 100]. Higher value => easier text.

    References:
        Muñoz, M., and J. Muñoz. "Legibilidad Mµ." Viña del Mar: CHL (2006).
    """
    n_words = len(n_chars_per_word)
    if n_words < 2:
        LOGGER.warning(
            "mu legibility index is undefined for texts with fewer than two words; "
            "returning 0.0"
        )
        return 0.0
    return (
        100
        * (n_words / (n_words - 1))
        * (statistics.mean(n_chars_per_word) / statistics.variance(n_chars_per_word))
    )


def perspicuity_index(n_syllables: int, n_words: int, n_sents: int) -> float:
    """
    Readability test for Spanish-language texts, whose value is in the range [0, 100];
    very similar to the Spanish-specific formulation of :func:`flesch_reading_ease()`,
    but included additionally since it's become a common readability standard.
    Higher value => easier text.

    References:
        Pazos, Francisco Szigriszt. Sistemas predictivos de legibilidad del mensaje
        escrito: fórmula de perspicuidad. Universidad Complutense de Madrid,
        Servicio de Reprografía, 1993.
    """
    return 206.835 - (n_words / n_sents) - (62.3 * (n_syllables / n_words))


def smog_index(n_polysyllable_words: int, n_sents: int) -> float:
    """
    Readability test commonly used in medical writing and the healthcare industry,
    whose value estimates the number of years of education required to understand a text
    similar to :func:`flesch_kincaid_grade_level()` and intended as a substitute for
    :func:`gunning_fog_index()`. Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/SMOG
    """
    if n_sents < 30:
        LOGGER.warning("SMOG index may be unreliable for n_sents < 30")
    return (1.0430 * sqrt(30 * n_polysyllable_words / n_sents)) + 3.1291


def wiener_sachtextformel(
    n_words: int,
    n_polysyllable_words: int,
    n_monosyllable_words: int,
    n_long_words: int,
    n_sents: int,
    *,
    variant: int = 1,
) -> float:
    """
    Readability test for German-language texts, whose value estimates the grade
    level required to understand a text. Higher value => more difficult text.

    References:
        https://de.wikipedia.org/wiki/Lesbarkeitsindex#Wiener_Sachtextformel
    """
    ms = 100 * n_polysyllable_words / n_words
    sl = n_words / n_sents
    iw = 100 * n_long_words / n_words
    es = 100 * n_monosyllable_words / n_words
    if variant == 1:
        return (0.1935 * ms) + (0.1672 * sl) + (0.1297 * iw) - (0.0327 * es) - 0.875
    elif variant == 2:
        return (0.2007 * ms) + (0.1682 * sl) + (0.1373 * iw) - 2.779
    elif variant == 3:
        return (0.2963 * ms) + (0.1905 * sl) - 1.1144
    elif variant == 4:
        return (0.2744 * ms) + (0.2656 * sl) - 1.693
    else:
        raise ValueError(errors.value_invalid_msg("variant", variant, [1, 2, 3, 4]))



import functools

import pytest
from spacy.tokens import Doc

from textacy import make_spacy_doc
from textacy.augmentation import augmenter, transforms


@pytest.fixture(scope="module")
def spacy_doc():
    text = (
        "Democrats might know that they stand against Trump's policies, but coming up with their own plan is harder than you think. "
        "For a long time, the party's top echelon has been captive to free trade orthodoxy. "
        "Since Bill Clinton, the theory of the case among the Democratic Party's elite has been the more globalization, the better — with mostly a deaf ear turned to the people and places most badly affected. "
        "Worse, their response to globalization's excesses has been: "
        "Here's a new trade deal, much better than the last one."
    )
    return make_spacy_doc(text, lang="en_core_web_sm")


@pytest.fixture(scope="module")
def example_augmenter():
    return augmenter.Augmenter(
        [transforms.swap_words, transforms.delete_words, transforms.swap_chars, transforms.delete_chars],
        num=None,
    )


class TestAugmenter:

    def test_bad_args(self):
        for num in [-1, 10, -0.1, 2.0, [0.5, 0.5]]:
            with pytest.raises(ValueError):
                _ = augmenter.Augmenter([transforms.substitute_word_synonyms], num=num)
        for tfs in [[], ["not callable obj"]]:
            with pytest.raises((ValueError, TypeError)):
                _ = augmenter.Augmenter(tfs, num=1)

    def test_good_args(self):
        for num in [None, 1, 2, 0.25, [0.5, 0.5]]:
            _ = augmenter.Augmenter(
                [transforms.substitute_chars, transforms.insert_chars], num=num,
            )
        test_tfs = [
            [transforms.substitute_chars],
            [lambda x: x],
            [transforms.delete_chars, functools.partial(transforms.substitute_chars, num=1)],
        ]
        for tfs in test_tfs:
            _ = augmenter.Augmenter(tfs, num=1)

    def test_apply_transforms(self, spacy_doc, example_augmenter):
        new_doc1 = example_augmenter.apply_transforms(spacy_doc, lang="en_core_web_sm")
        new_doc2 = example_augmenter.apply_transforms(spacy_doc, lang="en_core_web_sm")
        assert isinstance(new_doc1, Doc)
        assert new_doc1.text != spacy_doc.text
        assert new_doc1.text != new_doc2.text



import pytest

from textacy import preprocessing


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("$1.00 equals 100¢.", "_CUR_1.00 equals 100_CUR_."),
        ("How much is ¥100 in £?", "How much is _CUR_100 in _CUR_?"),
        ("My password is 123$abc฿.", "My password is 123_CUR_abc_CUR_."),
    ]
)
def test_replace_currency_symbols(text_in, text_out):
    assert preprocessing.replace.currency_symbols(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("Reach out at username@example.com.", "Reach out at _EMAIL_."),
        ("Click here: mailto:username@example.com.", "Click here: _EMAIL_."),
    ]
)
def test_replace_emails(text_in, text_out):
    assert preprocessing.replace.emails(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("ugh, it's raining *again* ☔", "ugh, it's raining *again* _EMOJI_"),
        ("✌ tests are passing ✌", "_EMOJI_ tests are passing _EMOJI_"),
    ]
)
def test_replace_emojis(text_in, text_out):
    assert preprocessing.replace.emojis(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("like omg it's #ThrowbackThursday", "like omg it's _TAG_"),
        ("#TextacyIn4Words: \"but it's honest work\"", "_TAG_: \"but it's honest work\""),
        ("wth twitter #ican'teven #why-even-try", "wth twitter _TAG_'teven _TAG_-even-try"),
        ("www.foo.com#fragment is not a hashtag", "www.foo.com#fragment is not a hashtag"),
    ]
)
def test_replace_hashtags(text_in, text_out):
    assert preprocessing.replace.hashtags(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        (
            "I owe $1,000.99 to 123 people for 2 +1 reasons.",
            "I owe $_NUMBER_ to _NUMBER_ people for _NUMBER_ _NUMBER_ reasons.",
        ),
    ]
)
def test_replace_numbers(text_in, text_out):
    assert preprocessing.replace.numbers(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        (
            "I can be reached at 555-123-4567 through next Friday.",
            "I can be reached at _PHONE_ through next Friday.",
        ),
    ]
)
def test_replace_phone_numbers(text_in, text_out):
    assert preprocessing.replace.phone_numbers(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        (
            "I learned everything I know from www.stackoverflow.com and http://wikipedia.org/ and Mom.",
            "I learned everything I know from _URL_ and _URL_ and Mom.",
        ),
    ]
)
def test_replace_urls(text_in, text_out):
    assert preprocessing.replace.urls(text_in) == text_out


@pytest.mark.parametrize(
    "text_in, text_out",
    [
        ("like omg it's @bjdewilde", "like omg it's _USER_"),
        ("@Real_Burton_DeWilde: definitely not a bot", "_USER_: definitely not a bot"),
        ("wth twitter @b.j.dewilde", "wth twitter _USER_.j.dewilde"),
        ("foo@bar.com is not a user handle", "foo@bar.com is not a user handle"),
    ]
)
def test_replace_user_handles(text_in, text_out):
    assert preprocessing.replace.user_handles(text_in) == text_out



import collections

import numpy as np
import pytest
from spacy.tokens import Doc

from textacy import Corpus
from textacy import load_spacy_lang
from textacy.datasets.capitol_words import CapitolWords

DATASET = CapitolWords()

pytestmark = pytest.mark.skipif(
    DATASET.filepath is None,
    reason="CapitolWords dataset must be downloaded before running tests",
)


@pytest.fixture(scope="module")
def en_core_web_sm():
    return load_spacy_lang("en_core_web_sm")


@pytest.fixture(scope="module")
def langs():
    return ["en_core_web_sm", load_spacy_lang("en_core_web_sm")]


@pytest.fixture(scope="module")
def corpus(en_core_web_sm):
    return Corpus(en_core_web_sm, data=DATASET.records(limit=5))


class TestCorpusInit:

    def test_init_lang(self, langs):
        for lang in langs:
            assert isinstance(Corpus(lang), Corpus)

    @pytest.mark.parametrize("lang", [b"en_core_web_sm", None])
    def test_init_bad_lang(self, lang):
        with pytest.raises(TypeError):
            Corpus(lang)

    def test_init_texts(self, en_core_web_sm):
        limit = 3
        texts = list(DATASET.texts(limit=limit))
        corpus = Corpus(en_core_web_sm, data=texts)
        assert len(corpus) == len(corpus.docs) == limit
        assert all(doc.vocab is corpus.spacy_lang.vocab for doc in corpus)
        assert all(text == doc.text for text, doc in zip(texts, corpus))

    def test_init_records(self, en_core_web_sm):
        limit = 3
        records = list(DATASET.records(limit=limit))
        corpus = Corpus(en_core_web_sm, data=records)
        assert len(corpus) == len(corpus.docs) == limit
        assert all(doc.vocab is corpus.spacy_lang.vocab for doc in corpus)
        assert all(record[0] == doc.text for record, doc in zip(records, corpus))
        assert all(record[1] == doc._.meta for record, doc in zip(records, corpus))

    def test_init_docs(self, en_core_web_sm):
        limit = 3
        texts = DATASET.texts(limit=limit)
        docs = [en_core_web_sm(text) for text in texts]
        corpus = Corpus(en_core_web_sm, data=docs)
        assert len(corpus) == len(corpus.docs) == limit
        assert all(doc.vocab is corpus.spacy_lang.vocab for doc in corpus)
        assert all(doc1 is doc2 for doc1, doc2 in zip(docs, corpus))

    def test_init_no_parser(self):
        spacy_lang = load_spacy_lang("en_core_web_sm", disable=("parser",))
        corpus = Corpus(spacy_lang, data=(spacy_lang("This is a sentence in a doc."),))
        assert len(corpus) == 1
        assert corpus.n_sents == 0


class TestCorpusDunder:

    def test_str(self, corpus):
        cstr = str(corpus)
        assert cstr.startswith("Corpus")
        assert all(f"{n}" in cstr for n in [corpus.n_docs, corpus.n_tokens])

    def test_len(self, corpus):
        assert isinstance(len(corpus), int)
        assert len(corpus) == len(corpus.docs) == corpus.n_docs

    def test_iter(self, corpus):
        assert isinstance(corpus, collections.abc.Iterable)

    def test_getitem(self, corpus):
        assert isinstance(corpus[0], Doc)
        assert all(isinstance(doc, Doc) for doc in corpus[0:2])

    def test_delitem(self, corpus):
        del corpus[-1]
        del corpus[-1:]
        with pytest.raises(TypeError):
            del corpus["foo"]


class TestCorpusProperties:

    def test_vectors(self, corpus):
        vectors = corpus.vectors
        assert isinstance(vectors, np.ndarray)
        assert len(vectors.shape) == 2
        assert vectors.shape[0] == len(corpus)

    def test_vector_norms(self, corpus):
        vector_norms = corpus.vector_norms
        assert isinstance(vector_norms, np.ndarray)
        assert len(vector_norms.shape) == 2
        assert vector_norms.shape[0] == len(corpus)


class TestCorpusMethods:

    # NOTE: this test must run before we add more docs that don't have valid metadata
    @pytest.mark.parametrize(
        "name, agg_func, exp_type",
        [
            ("date", min, str),
            ("congress", max, int),
            ("speaker_name", collections.Counter, dict),
        ],
    )
    def test_agg_metadata(self, name, agg_func, exp_type, corpus):
        result = corpus.agg_metadata(name, agg_func)
        assert isinstance(result, exp_type)

    def test_add(self, corpus, en_core_web_sm):
        datas = (
            "This is an english sentence.",
            ("This is an english sentence.", {"foo": "bar"}),
            en_core_web_sm("This is an english sentence."),
            ["This is one sentence.", "This is another sentence."],
            [("This is sentence #1.", {"foo": "bar"}), ("This is sentence #2.", {"bat": "baz"})],
            [en_core_web_sm("This is sentence #1"), en_core_web_sm("This is sentence #2")],
        )
        n_docs = corpus.n_docs
        for data in datas:
            corpus.add(data)
            assert corpus.n_docs > n_docs
            n_docs = corpus.n_docs

    @pytest.mark.xfail(reason="there seems to be bug in spacy for nprocess > 1 ...")
    def test_add_nprocess(self, corpus):
        datas = (
            ["This is one sentence.", "This is another sentence."],
            [("This is sentence #1.", {"foo": "bar"}), ("This is sentence #2.", {"bat": "baz"})],
        )
        n_docs = corpus.n_docs
        for data in datas:
            corpus.add(data, n_process=2)
            assert corpus.n_docs > n_docs
            n_docs = corpus.n_docs

    def test_add_typeerror(self, corpus):
        datas = (
            b"This is a byte string.",
            [b"This is a byte string.", b"This is another byte string."],
        )
        for data in datas:
            with pytest.raises(TypeError):
                corpus.add(data)

    def test_get(self, corpus):
        match_funcs = (
            lambda doc: True,
            lambda doc: doc._.meta.get("speaker_name") == "Bernie Sanders",
        )
        for match_func in match_funcs:
            assert len(list(corpus.get(match_func))) > 0
            assert len(list(corpus.get(match_func, limit=1))) == 1

    def test_remove(self, corpus):
        match_funcs = (
            lambda doc: doc._.meta.get("foo") == "bar",
            lambda doc: len(doc) < 10,
        )
        n_docs = corpus.n_docs
        for match_func in match_funcs[:1]:
            corpus.remove(match_func)
            assert corpus.n_docs < n_docs
            assert not any([match_func(doc) for doc in corpus])
            n_docs = corpus.n_docs

    def test_word_counts(self, corpus):
        abs_counts = corpus.word_counts(weighting="count", by="lower")
        rel_counts = corpus.word_counts(weighting="freq", by="lower")
        assert isinstance(abs_counts, dict)
        assert all(isinstance(count, int) for count in abs_counts.values())
        assert min(abs_counts.values()) > 0
        assert isinstance(rel_counts, dict)
        assert all(isinstance(count, float) for count in rel_counts.values())
        assert min(rel_counts.values()) > 0 and max(rel_counts.values()) <= 1

    def test_word_counts_error(self, corpus):
        with pytest.raises(ValueError):
            corpus.word_counts(weighting="foo")

    def test_word_doc_counts(self, corpus):
        abs_counts = corpus.word_doc_counts(weighting="count", by="lower")
        rel_counts = corpus.word_doc_counts(weighting="freq", by="lower")
        inv_counts = corpus.word_doc_counts(weighting="idf", by="lower")
        assert isinstance(abs_counts, dict)
        assert all(isinstance(count, int) for count in abs_counts.values())
        assert min(abs_counts.values()) > 0
        assert isinstance(rel_counts, dict)
        assert all(isinstance(count, float) for count in rel_counts.values())
        assert min(rel_counts.values()) > 0 and max(rel_counts.values()) <= 1
        assert isinstance(inv_counts, dict)
        assert min(inv_counts.values()) > 0

    def test_word_doc_counts_error(self, corpus):
        with pytest.raises(ValueError):
            corpus.word_doc_counts(weighting="foo")

    def test_corpus_save_and_load(self, corpus, tmpdir, en_core_web_sm):
        filepath = str(tmpdir.join("test_corpus_save_and_load.bin"))
        corpus.save(filepath)
        loaded_corpus = Corpus.load(en_core_web_sm, filepath)
        assert isinstance(loaded_corpus, Corpus)
        assert len(loaded_corpus) == len(corpus)
        assert loaded_corpus.spacy_lang.meta == corpus.spacy_lang.meta
        assert loaded_corpus.spacy_lang.pipe_names == corpus.spacy_lang.pipe_names
        assert corpus[0].user_data == loaded_corpus[0].user_data



"""
:mod:`textacy.io.spacy`: Functions for reading from and writing to disk spacy documents
in either pickle or binary format. Be warned: Both formats have pros and cons.
"""
from __future__ import annotations

import pickle
from typing import Iterable, Optional

from spacy.tokens import Doc, DocBin

from .. import errors, spacier, types
from . import utils as io_utils


def read_spacy_docs(
    filepath: types.PathLike,
    *,
    format: str = "binary",
    lang: Optional[types.LangLike] = None,
) -> Iterable[Doc]:
    """
    Read the contents of a file at ``filepath``, written in binary or pickle format.

    Args:
        filepath: Path to file on disk from which data will be read.
        format ({"binary", "pickle"}): Format of the data that was written to disk.
            If "binary", uses :class:`spacy.tokens.DocBin` to deserialie data;
            if "pickle", uses python's stdlib ``pickle``.

            .. warning:: Docs written in pickle format were saved all together
               as a list, which means they're all loaded into memory at once
               before streaming one by one. Mind your RAM usage, especially when
               reading many docs!

        lang: Language with which spaCy originally processed docs, represented as
            the full name of or path on disk to the pipeline, or an already instantiated
            pipeline instance.
            Note that this is only required when ``format`` is "binary".

    Yields:
        Next deserialized document.

    Raises:
        ValueError: if format is not "binary" or "pickle", or if ``lang`` is None
            when ``format="binary"``
    """
    if format == "binary":
        if lang is None:
            raise ValueError(
                "lang=None is invalid. When format='binary', a `spacy.Language` "
                "(well, its associated `spacy.Vocab`) is required to deserialize "
                "the binary data. Note that this should be the same language pipeline "
                "used when processing the original docs!"
            )
        else:
            lang = spacier.utils.resolve_langlike(lang)
        docbin = DocBin().from_disk(filepath)
        for doc in docbin.get_docs(lang.vocab):
            yield doc

    elif format == "pickle":
        with io_utils.open_sesame(filepath, mode="rb") as f:
            for spacy_doc in pickle.load(f):
                yield spacy_doc

    else:
        raise ValueError(
            errors.value_invalid_msg("format", format, {"binary", "pickle"})
        )


def write_spacy_docs(
    data: Doc | Iterable[Doc],
    filepath: types.PathLike,
    *,
    make_dirs: bool = False,
    format: str = "binary",
    attrs: Optional[Iterable[str]] = None,
    store_user_data: bool = False,
) -> None:
    """
    Write one or more ``Doc`` s to disk at ``filepath`` in binary or pickle format.

    Args:
        data: A single ``Doc`` or a sequence of ``Doc`` s to write to disk.
        filepath: Path to file on disk to which data will be written.
        make_dirs: If True, automatically create (sub)directories
            if not already present in order to write ``filepath``.
        format ({"pickle", "binary"}): Format of the data written to disk.
            If "binary", uses :class:`spacy.tokens.DocBin` to serialie data;
            if "pickle", uses python's stdlib ``pickle``.

            .. warning:: When writing docs in pickle format, all the docs in ``data``
               must be saved as a list, which means they're all loaded into memory.
               Mind your RAM usage, especially when writing many docs!

        attrs: List of attributes to serialize if ``format`` is "binary". If None,
            spaCy's default values are used; see here: https://spacy.io/api/docbin#init
        store_user_data: If True, write :attr`Doc.user_data` and the values of custom
            extension attributes to disk; otherwise, don't.

    Raises:
        ValueError: if format is not "binary" or "pickle"
    """
    if isinstance(data, Doc):
        data = [data]
    if format == "binary":
        kwargs = {"docs": data, "store_user_data": store_user_data}
        if attrs is not None:
            kwargs["attrs"] = list(attrs)
        docbin = DocBin(**kwargs)
        docbin.to_disk(filepath)
    elif format == "pickle":
        if store_user_data is False:
            data = _clear_docs_user_data(data)
        with io_utils.open_sesame(filepath, mode="wb", make_dirs=make_dirs) as f:
            pickle.dump(list(data), f, protocol=-1)
    else:
        raise ValueError(
            errors.value_invalid_msg("format", format, {"binary", "pickle"})
        )


def _clear_docs_user_data(docs: Iterable[Doc]) -> Iterable[Doc]:
    # TODO: figure out if/how to clear out custom doc extension values
    for doc in docs:
        doc.user_data.clear()
        yield doc



from typing import Any, Dict

from spacy.tokens import Doc

from . import acros, basics, keyterms, kwic, matches, triples
from .. import errors


def extract_keyterms(doc: Doc, method: str, **kwargs):
    """
    Extract keyterms from a document using one of several different methods.
    For full detail, see the underlying functions listed below.

    See Also:
        - :func:`textacy.extract.keyterms.scake()`
        - :func:`textacy.extract.keyterms.sgrank()`
        - :func:`textacy.extract.keyterms.textrank()`
        - :func:`textacy.extract.keyterms.yake()`
    """
    if method == "scake":
        return keyterms.scake(doc, **kwargs)
    elif method == "sgrank":
        return keyterms.sgrank(doc, **kwargs)
    elif method == "textrank":
        return keyterms.textrank(doc, **kwargs)
    elif method == "yake":
        return keyterms.yake(doc, **kwargs)
    else:
        raise ValueError(
            errors.value_invalid_msg(
                "method", method, {"scake", "sgrank", "textrank", "yake"}
            )
        )


def get_doc_extensions() -> Dict[str, Dict[str, Any]]:
    """
    Get :mod:`textacy.extract` custom property and method doc extensions
    that can be set on or removed from the global :class:`spacy.tokens.Doc`.
    """
    return _DOC_EXTENSIONS


def set_doc_extensions():
    """
    Set :mod:`textacy.extract` custom property and method doc extensions
    on the global :class:`spacy.tokens.Doc`.
    """
    for name, kwargs in get_doc_extensions().items():
        if not Doc.has_extension(name):
            Doc.set_extension(name, **kwargs)


def remove_doc_extensions():
    """
    Remove :mod:`textacy.extract` custom property and method doc extensions
    from the global :class:`spacy.tokens.Doc`.
    """
    for name in get_doc_extensions().keys():
        _ = Doc.remove_extension(name)


_DOC_EXTENSIONS: Dict[str, Dict[str, Any]] = {
    "extract_words": {"method": basics.words},
    "extract_ngrams": {"method": basics.ngrams},
    "extract_entities": {"method": basics.entities},
    "extract_noun_chunks": {"method": basics.noun_chunks},
    "extract_terms": {"method": basics.terms},
    "extract_token_matches": {"method": matches.token_matches},
    "extract_regex_matches": {"method": matches.regex_matches},
    "extract_subject_verb_object_triples": {
        "method": triples.subject_verb_object_triples
    },
    "extract_semistructured_statements": {"method": triples.semistructured_statements},
    "extract_direct_quotations": {"method": triples.direct_quotations},
    "extract_acronyms": {"method": acros.acronyms},
    "extract_acronyms_and_definitions": {"method": acros.acronyms_and_definitions},
    "extract_keyword_in_context": {"method": kwic.keyword_in_context},
    "extract_keyterms": {"method": extract_keyterms},
}



"""
Triples
-------

:mod:`textacy.extract.triples`: Extract structured triples from a document or sentence
through rule-based pattern-matching of the annotated tokens.
"""
from __future__ import annotations

import collections
from operator import attrgetter
from typing import Iterable, List, Optional, Pattern, Tuple

from cytoolz import itertoolz
from spacy.symbols import (
    AUX, VERB,
    agent, attr, aux, auxpass, csubj, csubjpass, dobj, neg, nsubj, nsubjpass, obj, pobj, xcomp,
)
from spacy.tokens import Doc, Span, Token

from . import matches
from .. import constants, types, utils


_NOMINAL_SUBJ_DEPS = {nsubj, nsubjpass}
_CLAUSAL_SUBJ_DEPS = {csubj, csubjpass}
_ACTIVE_SUBJ_DEPS = {csubj, nsubj}
_VERB_MODIFIER_DEPS = {aux, auxpass, neg}

SVOTriple: Tuple[List[Token], List[Token], List[Token]] = collections.namedtuple(
    "SVOTriple", ["subject", "verb", "object"]
)
SSSTriple: Tuple[List[Token], List[Token], List[Token]] = collections.namedtuple(
    "SSSTriple", ["entity", "cue", "fragment"]
)
DQTriple: Tuple[List[Token], List[Token], Span] = collections.namedtuple(
    "DQTriple", ["speaker", "cue", "content"]
)


def subject_verb_object_triples(doclike: types.DocLike) -> Iterable[SVOTriple]:
    """
    Extract an ordered sequence of subject-verb-object triples from a document
    or sentence.

    Args:
        doclike

    Yields:
        Next SVO triple as (subject, verb, object), in approximate order of appearance.
    """
    if isinstance(doclike, Span):
        sents = [doclike]
    else:
        sents = doclike.sents

    for sent in sents:
        # connect subjects/objects to direct verb heads
        # and expand them to include conjuncts, compound nouns, ...
        verb_sos = collections.defaultdict(lambda: collections.defaultdict(set))
        for tok in sent:
            head = tok.head
            # ensure entry for all verbs, even if empty
            # to catch conjugate verbs without direct subject/object deps
            if tok.pos == VERB:
                _ = verb_sos[tok]
            # nominal subject of active or passive verb
            if tok.dep in _NOMINAL_SUBJ_DEPS:
                if head.pos == VERB:
                    verb_sos[head]["subjects"].update(expand_noun(tok))
            # clausal subject of active or passive verb
            elif tok.dep in _CLAUSAL_SUBJ_DEPS:
                if head.pos == VERB:
                    verb_sos[head]["subjects"].update(tok.subtree)
            # nominal direct object of transitive verb
            elif tok.dep == dobj:
                if head.pos == VERB:
                    verb_sos[head]["objects"].update(expand_noun(tok))
            # prepositional object acting as agent of passive verb
            elif tok.dep == pobj:
                if head.dep == agent and head.head.pos == VERB:
                    verb_sos[head.head]["objects"].update(expand_noun(tok))
            # open clausal complement, but not as a secondary predicate
            elif tok.dep == xcomp:
                if (
                    head.pos == VERB
                    and not any(child.dep == dobj for child in head.children)
                ):
                    # TODO: just the verb, or the whole tree?
                    # verb_sos[verb]["objects"].update(expand_verb(tok))
                    verb_sos[head]["objects"].update(tok.subtree)
        # fill in any indirect relationships connected via verb conjuncts
        for verb, so_dict in verb_sos.items():
            conjuncts = verb.conjuncts
            if so_dict.get("subjects"):
                for conj in conjuncts:
                    conj_so_dict = verb_sos.get(conj)
                    if conj_so_dict and not conj_so_dict.get("subjects"):
                        conj_so_dict["subjects"].update(so_dict["subjects"])
            if not so_dict.get("objects"):
                so_dict["objects"].update(
                    obj
                    for conj in conjuncts
                    for obj in verb_sos.get(conj, {}).get("objects", [])
                )
        # expand verbs and restructure into svo triples
        for verb, so_dict in verb_sos.items():
            if so_dict["subjects"] and so_dict["objects"]:
                yield SVOTriple(
                    subject=sorted(so_dict["subjects"], key=attrgetter("i")),
                    verb=sorted(expand_verb(verb), key=attrgetter("i")),
                    object=sorted(so_dict["objects"], key=attrgetter("i")),
                )


def semistructured_statements(
    doclike: types.DocLike,
    *,
    entity: str | Pattern,
    cue: str,
    fragment_len_range: Optional[Tuple[Optional[int], Optional[int]]] = None,
) -> Iterable[SSSTriple]:
    """
    Extract "semi-structured statements" from a document as a sequence of
    (entity, cue, fragment) triples.

    Args:
        doclike
        entity: Noun or noun phrase of interest expressed as a regular expression
            pattern string (e.g. ``"[Gg]lobal [Ww]arming"``) or compiled object
            (e.g. ``re.compile("global warming", re.IGNORECASE)``).
        cue: Verb lemma with which ``entity`` is associated (e.g. "be", "have", "say").
        fragment_len_range: Filter statements to those whose fragment length in tokens
            is within the specified [low, high) interval. Both low and high values
            must be specified, but a null value for either is automatically replaced
            by safe default values. None (default) skips filtering by fragment length.

    Yields:
        Next matching triple, consisting of (entity, cue, fragment),
        in order of appearance.

    Notes:
        Inspired by N. Diakopoulos, A. Zhang, A. Salway. Visual Analytics of
        Media Frames in Online News and Blogs. IEEE InfoVis Workshop on Text
        Visualization. October, 2013.

        Which itself was inspired by by Salway, A.; Kelly, L.; Skadiņa, I.; and
        Jones, G. 2010. Portable Extraction of Partially Structured Facts from
        the Web. In Proc. ICETAL 2010, LNAI 6233, 345-356. Heidelberg, Springer.
    """
    if fragment_len_range is not None:
        fragment_len_range = utils.validate_and_clip_range(
            fragment_len_range, (1, 1000), int
        )
    for entity_cand in matches.regex_matches(doclike, entity, alignment_mode="strict"):
        # is the entity candidate a nominal subject?
        if entity_cand.root.dep in _NOMINAL_SUBJ_DEPS:
            cue_cand = entity_cand.root.head
            # is the cue candidate a verb with matching lemma?
            if cue_cand.pos in {VERB, AUX} and cue_cand.lemma_ == cue:
                frag_cand = None
                for tok in cue_cand.children:
                    if (
                        tok.dep in {attr, dobj, obj}
                        or tok.dep_ == "dative"
                        or (
                            tok.dep == xcomp
                            and not any(child.dep == dobj for child in cue_cand.children)
                        )
                    ):
                        subtoks = list(tok.subtree)
                        if (
                            fragment_len_range is None
                            or fragment_len_range[0] <= len(subtoks) < fragment_len_range[1]
                        ):
                            frag_cand = subtoks
                            break
                if frag_cand is not None:
                    yield SSSTriple(
                        entity=list(entity_cand),
                        cue=sorted(expand_verb(cue_cand), key=attrgetter("i")),
                        fragment=sorted(frag_cand, key=attrgetter("i")),
                    )


def direct_quotations(doc: Doc) -> Iterable[DQTriple]:
    """
    Extract direct quotations with an attributable speaker from a document
    using simple rules and patterns. Does not extract indirect or mixed quotations!

    Args:
        doc

    Yields:
        Next direct quotation in ``doc`` as a (speaker, cue, content) triple.

    Notes:
        Loosely inspired by Krestel, Bergler, Witte. "Minding the Source: Automatic
        Tagging of Reported Speech in Newspaper Articles".
    """
    # TODO: train a model to do this instead, maybe similar to entity recognition
    try:
        _reporting_verbs = constants.REPORTING_VERBS[doc.lang_]
    except KeyError:
        raise ValueError(
            f"direct quotation extraction is not implemented for lang='{doc.lang_}', "
            f"only {sorted(constants.REPORTING_VERBS.keys())}"
        )
    qtok_idxs = [tok.i for tok in doc if tok.is_quote]
    if len(qtok_idxs) % 2 != 0:
        raise ValueError(
            f"{len(qtok_idxs)} quotation marks found, indicating an unclosed quotation; "
            "given the limitations of this method, it's safest to bail out "
            "rather than guess which quotation is unclosed"
        )
    qtok_pair_idxs = list(itertoolz.partition(2, qtok_idxs))
    for qtok_start_idx, qtok_end_idx in qtok_pair_idxs:
        content = doc[qtok_start_idx : qtok_end_idx + 1]
        cue = None
        speaker = None
        # filter quotations by content
        if (
            # quotations should have at least a couple tokens
            # excluding the first/last quotation mark tokens
            len(content) < 4
            # filter out titles of books and such, if possible
            or all(
                tok.is_title
                for tok in content
                # if tok.pos in {NOUN, PROPN}
                if not (tok.is_punct or tok.is_stop)
            )
            # TODO: require closing punctuation before the quotation mark?
            # content[-2].is_punct is False
        ):
            continue
        # get window of adjacent/overlapping sentences
        window_sents = (
            sent
            for sent in doc.sents
            # these boundary cases are a subtle bit of work...
            if (
                (sent.start < qtok_start_idx and sent.end >= qtok_start_idx - 1)
                or (sent.start <= qtok_end_idx + 1 and sent.end > qtok_end_idx)
            )
        )
        # get candidate cue verbs in window
        cue_cands = [
            tok
            for sent in window_sents
            for tok in sent
            if (
                tok.pos == VERB
                and tok.lemma_ in _reporting_verbs
                # cue verbs must occur *outside* any quotation content
                and not any(
                    qts_idx <= tok.i <= qte_idx
                    for qts_idx, qte_idx in qtok_pair_idxs
                )
            )
        ]
        # sort candidates by proximity to quote content
        cue_cands = sorted(
            cue_cands,
            key=lambda cc: min(abs(cc.i - qtok_start_idx), abs(cc.i - qtok_end_idx)),
        )
        for cue_cand in cue_cands:
            if cue is not None:
                break
            for speaker_cand in cue_cand.children:
                if speaker_cand.dep in _ACTIVE_SUBJ_DEPS:
                    cue = expand_verb(cue_cand)
                    speaker = expand_noun(speaker_cand)
                    break
        if content and cue and speaker:
            yield DQTriple(
                speaker=sorted(speaker, key=attrgetter("i")),
                cue=sorted(cue, key=attrgetter("i")),
                content=content,
            )


def expand_noun(tok: Token) -> List[Token]:
    """Expand a noun token to include all associated conjunct and compound nouns."""
    tok_and_conjuncts = [tok] + list(tok.conjuncts)
    compounds = [
        child
        for tc in tok_and_conjuncts
        for child in tc.children
        # TODO: why doesn't compound import from spacy.symbols?
        if child.dep_ == "compound"
    ]
    return tok_and_conjuncts + compounds


def expand_verb(tok: Token) -> List[Token]:
    """Expand a verb token to include all associated auxiliary and negation tokens."""
    verb_modifiers = [
        child for child in tok.children if child.dep in _VERB_MODIFIER_DEPS
    ]
    return [tok] + verb_modifiers



from __future__ import annotations

import collections
import itertools
from operator import itemgetter
from typing import (
    Callable,
    Collection,
    Counter,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
)

import networkx as nx
from cytoolz import itertoolz
from spacy.tokens import Doc, Token

from ... import utils
from .. import utils as ext_utils


def scake(
    doc: Doc,
    *,
    normalize: Optional[str | Callable[[Token], str]] = "lemma",
    include_pos: Optional[str | Collection[str]] = ("NOUN", "PROPN", "ADJ"),
    topn: int | float = 10,
) -> List[Tuple[str, float]]:
    """
    Extract key terms from a document using the sCAKE algorithm.

    Args:
        doc: spaCy ``Doc`` from which to extract keyterms. Must be sentence-segmented;
            optionally POS-tagged.
        normalize: If "lemma", lemmatize terms; if "lower", lowercase terms; if None,
            use the form of terms as they appeared in ``doc``; if a callable,
            must accept a ``Token`` and return a str,
            e.g. :func:`textacy.spacier.utils.get_normalized_text()`.
        include_pos: One or more POS tags with which to filter for good candidate keyterms.
            If None, include tokens of all POS tags
            (which also allows keyterm extraction from docs without POS-tagging.)
        topn: Number of top-ranked terms to return as key terms.
            If an integer, represents the absolute number; if a float, value
            must be in the interval (0.0, 1.0], which is converted to an int by
            ``int(round(len(candidates) * topn))``

    Returns:
        Sorted list of top ``topn`` key terms and their corresponding scores.

    References:
        Duari, Swagata & Bhatnagar, Vasudha. (2018). sCAKE: Semantic Connectivity
        Aware Keyword Extraction. Information Sciences. 477.
        https://arxiv.org/abs/1811.10831v1
    """
    # validate / transform args
    include_pos = utils.to_collection(include_pos, str, set)
    if isinstance(topn, float):
        if not 0.0 < topn <= 1.0:
            raise ValueError(
                "topn={} is invalid; "
                "must be an int, or a float between 0.0 and 1.0".format(topn)
            )

    # bail out on empty docs
    if not doc:
        return []

    # build up a graph of good words, edges weighting by adjacent sentence co-occurrence
    cooc_mat: Counter[Tuple[str, str]] = collections.Counter()
    # handle edge case where doc only has 1 sentence
    n_sents = itertoolz.count(doc.sents)
    for window_sents in itertoolz.sliding_window(min(2, n_sents), doc.sents):
        if n_sents == 1:
            window_sents = (window_sents[0], [])
        window_words: Iterable[str] = (
            word
            for word in itertoolz.concat(window_sents)
            if not (word.is_stop or word.is_punct or word.is_space)
            and (not include_pos or word.pos_ in include_pos)
        )
        window_words = ext_utils.terms_to_strings(window_words, normalize)
        cooc_mat.update(
            w1_w2
            for w1_w2 in itertools.combinations(sorted(window_words), 2)
            if w1_w2[0] != w1_w2[1]
        )
    # doc doesn't have any valid words...
    if not cooc_mat:
        return []

    graph = nx.Graph()
    graph.add_edges_from(
        (w1, w2, {"weight": weight}) for (w1, w2), weight in cooc_mat.items()
    )

    word_scores = _compute_word_scores(doc, graph, cooc_mat, normalize)
    if not word_scores:
        return []

    # generate a list of candidate terms
    candidates = _get_candidates(doc, normalize, include_pos)
    if isinstance(topn, float):
        topn = int(round(len(set(candidates)) * topn))
    # rank candidates by aggregating constituent word scores
    candidate_scores = {
        " ".join(candidate): sum(word_scores.get(word, 0.0) for word in candidate)
        for candidate in candidates
    }
    sorted_candidate_scores = sorted(
        candidate_scores.items(), key=itemgetter(1, 0), reverse=True
    )
    return ext_utils.get_filtered_topn_terms(
        sorted_candidate_scores, topn, match_threshold=0.8
    )


def _compute_word_scores(
    doc: Doc,
    graph: nx.Graph,
    cooc_mat: Dict[Tuple[str, str], int],
    normalize: Optional[str | Callable[[Token], str]],
) -> Dict[str, float]:
    word_strs: List[str] = list(graph.nodes())
    # "level of hierarchy" component
    max_truss_levels = _compute_node_truss_levels(graph)
    max_truss_level = max(max_truss_levels.values())
    # check for edge case when all word scores would be zero / undefined
    if not max_truss_level:
        return {}

    # "semantic strength of a word" component
    sem_strengths = {
        w: sum(
            cooc_mat[tuple(sorted([w, nbr]))] * max_truss_levels[nbr]
            for nbr in graph.neighbors(w)
        )
        for w in word_strs
    }
    # "semantic connectivity" component
    sem_connectivities = {
        w: len(set(max_truss_levels[nbr] for nbr in graph.neighbors(w)))
        / max_truss_level
        for w in word_strs
    }
    # "positional weight" component
    word_pos = collections.defaultdict(float)
    for word, word_str in zip(doc, ext_utils.terms_to_strings(doc, normalize)):
        word_pos[word_str] += 1 / (word.i + 1)
    return {
        w: word_pos[w] * max_truss_levels[w] * sem_strengths[w] * sem_connectivities[w]
        for w in word_strs
    }


def _get_candidates(
    doc: Doc,
    normalize: Optional[str | Callable[[Token], str]],
    include_pos: Set[str],
) -> Set[Tuple[str, ...]]:
    """
    Get a set of candidate terms to be scored by joining the longest
    subsequences of valid words -- non-stopword and non-punct, filtered to
    nouns, proper nouns, and adjectives if ``doc`` is POS-tagged -- then
    normalized into strings.
    """

    def _is_valid_tok(tok):
        return not (tok.is_stop or tok.is_punct or tok.is_space) and (
            not include_pos or tok.pos_ in include_pos
        )

    candidates = ext_utils.get_longest_subsequence_candidates(doc, _is_valid_tok)
    return {
        tuple(ext_utils.terms_to_strings(candidate, normalize))
        for candidate in candidates
    }


def _compute_node_truss_levels(graph: nx.Graph) -> Dict[str, int]:
    """
    Reference:
        Burkhardt, Paul & Faber, Vance & G. Harris, David. (2018).
        Bounds and algorithms for $k$-truss.
        https://arxiv.org/abs/1806.05523v1
    """
    max_edge_ks = {}
    is_removed = collections.defaultdict(int)
    triangle_counts = {
        edge: len(set(graph.neighbors(edge[0])) & set(graph.neighbors(edge[1])))
        for edge in graph.edges()
    }
    # rather than iterating over all theoretical values of k
    # let's break out early once all edges have been removed
    # max_edge_k = math.ceil(math.sqrt(len(triangle_counts)))
    # for k in range(1, max_edge_k):
    k = 1
    while True:
        to_remove = collections.deque(
            edge
            for edge, tcount in triangle_counts.items()
            if tcount < k and not is_removed[edge]
        )
        while to_remove:
            edge = to_remove.popleft()
            is_removed[edge] = 1
            for nbr in set(graph.neighbors(edge[0])) & set(graph.neighbors(edge[1])):
                for node in edge:
                    nbr_edge = (node, nbr)
                    try:
                        triangle_counts[nbr_edge] -= 1
                    except KeyError:
                        # oops, gotta reverse the node ordering on this edge
                        nbr_edge = (nbr, node)
                        triangle_counts[nbr_edge] -= 1
                    if triangle_counts[nbr_edge] == k - 1:
                        to_remove.append(nbr_edge)
                        is_removed[nbr_edge] = 1
            max_edge_ks[edge] = k - 1
        # here's where we break out early, if possible
        if len(is_removed) == len(triangle_counts):
            break
        else:
            k += 1
    max_node_ks = {
        node: max(k for edge, k in max_edge_ks.items() if node in edge)
        for node in graph.nodes()
    }
    return max_node_ks



import pytest

import textacy
from textacy import extract


@pytest.fixture(scope="module")
def doc():
    nlp = textacy.load_spacy_lang("en_core_web_sm")
    text = (
        "Many years later, as he faced the firing squad, Colonel Aureliano Buendía was "
        "to remember that distant afternoon when his father took him to discover ice. "
        "At that time Macondo was a village of twenty adobe houses, built on the bank "
        "of a river of clear water that ran along a bed of polished stones, which were "
        "white and enormous, like prehistoric eggs. The world was so recent that many "
        "things lacked names, and in order to indicate them it was necessary to point."
    )
    return nlp(text)


def test_extensions_exist(doc):
    for name in extract.extensions.get_doc_extensions().keys():
        assert doc.has_extension(name)


@pytest.mark.parametrize(
    "ext_name, kwargs",
    [
        ("extract_words", {}),
        ("extract_ngrams", {"n": 2}),
        ("extract_entities", {}),
        ("extract_noun_chunks", {}),
        ("extract_terms", {"ngs": 2, "ents": True, "ncs": True}),
        ("extract_token_matches", {"patterns": [{"POS": "NOUN"}]}),
        ("extract_regex_matches", {"pattern": "[Mm]any"}),
        ("extract_subject_verb_object_triples", {}),
        ("extract_semistructured_statements", {"entity": "Macondo", "cue": "is"}),
        ("extract_direct_quotations", {}),
        ("extract_acronyms", {}),
        ("extract_keyword_in_context", {"keyword": "Macondo"}),
    ]
)
def test_extensions_match(doc, ext_name, kwargs):
    ext = getattr(doc._, ext_name)
    func = extract.extensions.get_doc_extensions()[ext_name]["method"]
    ext_val = ext(**kwargs)
    func_val = func(doc, **kwargs)
    if isinstance(func_val, dict):
        assert ext_val == func_val
    else:
        assert list(ext_val) == list(func_val)



"""
Keyterms
--------

:mod:`textacy.extract.keyterms`: Extract keyterms from documents using a variety of
rule-based algorithms.
"""
from .scake import scake
from .sgrank import sgrank
from .textrank import textrank
from .yake import yake



import pytest
import scipy.sparse as sp

import textacy
from textacy import extract, representations


@pytest.fixture(scope="module")
def tokenized_docs():
    texts = [
        "Mary had a little lamb. Its fleece was white as snow.",
        "Everywhere that Mary went the lamb was sure to go.",
        "It followed her to school one day, which was against the rule.",
        "It made the children laugh and play to see a lamb at school.",
        "And so the teacher turned it out, but still it lingered near.",
        "It waited patiently about until Mary did appear.",
        "Why does the lamb love Mary so? The eager children cry.",
        "Mary loves the lamb, you know, the teacher did reply.",
    ]
    nlp = textacy.load_spacy_lang("en_core_web_sm")
    docs = list(nlp.pipe(texts))
    tokenized_docs = [
        [term.text.lower() for term in extract.terms(doc, ngs=1)] for doc in docs
    ]
    return tokenized_docs


@pytest.fixture(scope="module")
def groups():
    return ["a", "b", "c", "a", "b", "c", "a", "b"]


def test_build_doc_term_matrix(tokenized_docs):
    result = representations.build_doc_term_matrix(tokenized_docs)
    assert isinstance(result, tuple) and len(result) == 2
    dtm, vocab_terms = result
    assert isinstance(dtm, sp.csr_matrix)
    assert isinstance(vocab_terms, dict)


def test_build_grp_term_matrix(tokenized_docs, groups):
    result = representations.build_grp_term_matrix(tokenized_docs, groups)
    assert isinstance(result, tuple) and len(result) == 3
    gtm, vocab_terms, vocab_grps = result
    assert isinstance(gtm, sp.csr_matrix)
    assert isinstance(vocab_terms, dict)
    assert isinstance(vocab_grps, dict)



import pytest
import numpy as np
from scipy.sparse import coo_matrix

import textacy
from textacy import extract, representations


@pytest.fixture(scope="module")
def tokenized_docs():
    texts = [
        "Mary had a little lamb. Its fleece was white as snow.",
        "Everywhere that Mary went the lamb was sure to go.",
        "It followed her to school one day, which was against the rule.",
        "It made the children laugh and play to see a lamb at school.",
        "And so the teacher turned it out, but still it lingered near.",
        "It waited patiently about until Mary did appear.",
        "Why does the lamb love Mary so? The eager children cry.",
        "Mary loves the lamb, you know, the teacher did reply.",
    ]
    nlp = textacy.load_spacy_lang("en_core_web_sm")
    docs = list(nlp.pipe(texts))
    tokenized_docs = [
        [term.text.lower() for term in extract.terms(doc, ngs=1)] for doc in docs
    ]
    return tokenized_docs


@pytest.fixture(scope="module")
def groups():
    return ["a", "b", "c", "a", "b", "c", "a", "b"]


@pytest.fixture(scope="module")
def vectorizer_and_dtm(tokenized_docs):
    vectorizer = representations.Vectorizer(
        tf_type="linear",
        norm=None,
        min_df=1,
        max_df=1.0,
        max_n_terms=None,
    )
    doc_term_matrix = vectorizer.fit_transform(tokenized_docs)
    return vectorizer, doc_term_matrix


@pytest.fixture(scope="module")
def grp_vectorizer_and_gtm(tokenized_docs, groups):
    grp_vectorizer = representations.GroupVectorizer(
        tf_type="linear",
        norm=None,
        min_df=1,
        max_df=1.0,
        max_n_terms=None,
    )
    grp_term_matrix = grp_vectorizer.fit_transform(tokenized_docs, groups)
    return grp_vectorizer, grp_term_matrix


@pytest.fixture(scope="module")
def grp_vectorizer_and_gtm_2(tokenized_docs, groups):
    grp_vectorizer = representations.GroupVectorizer(
        tf_type="bm25",
        norm=None,
        dl_type="sqrt",
        min_df=1,
        max_df=1.0,
        max_n_terms=None,
    )
    grp_term_matrix = grp_vectorizer.fit_transform(tokenized_docs, groups)
    return grp_vectorizer, grp_term_matrix


@pytest.fixture(scope="module")
def lamb_and_children_idxs(vectorizer_and_dtm):
    vec, _ = vectorizer_and_dtm
    idx_lamb = [id_ for term, id_ in vec.vocabulary_terms.items() if term == "lamb"][0]
    idx_children = [id_ for term, id_ in vec.vocabulary_terms.items() if term == "children"][0]
    return idx_lamb, idx_children


class TestVectorizer:

    @pytest.mark.parametrize(
        "kwargs",
        [
            {"tf_type": "linear"},
            {"tf_type": "linear", "idf_type": "standard"},
            {"tf_type": "linear", "idf_type": "standard", "dl_type": "linear"},
            {"tf_type": "linear", "idf_type": "standard", "dl_type": "linear", "norm": "l1"},
            {"tf_type": "linear", "idf_type": "standard", "dl_type": "linear", "norm": "l2"},
            {"tf_type": "sqrt", "idf_type": "smooth", "dl_type": "sqrt"},
            {"tf_type": "log", "idf_type": "bm25", "dl_type": "log"},
            {"tf_type": "binary", "idf_type": None, "dl_type": None},
        ],
    )
    def test_weighting_combos(self, kwargs, tokenized_docs):
        vectorizer = representations.Vectorizer(**kwargs)
        _ = vectorizer.fit(tokenized_docs)
        _ = vectorizer.weighting

    @pytest.mark.parametrize(
        "kwargs",
        [
            {"min_df": -1},
            {"max_df": -1},
            {"max_n_terms": -1},
            {"vocabulary_terms": "foo bar bat baz"},
        ]
    )
    def test_bad_init_params(self, kwargs):
        with pytest.raises(ValueError):
            _ = representations.Vectorizer(**kwargs)

    def test_id_to_term(self, vectorizer_and_dtm):
        vectorizer, _ = vectorizer_and_dtm
        assert isinstance(vectorizer.id_to_term, dict)
        assert all(isinstance(key, int) for key in vectorizer.id_to_term.keys())
        assert all(
            isinstance(val, str) for val in vectorizer.id_to_term.values()
        )
        assert len(vectorizer.id_to_term) == len(vectorizer.vocabulary_terms)

    def test_terms_list(self, vectorizer_and_dtm):
        vectorizer, dtm = vectorizer_and_dtm
        assert isinstance(vectorizer.terms_list, list)
        assert isinstance(vectorizer.terms_list[0], str)
        assert len(vectorizer.terms_list) == len(vectorizer.vocabulary_terms)
        assert len(vectorizer.terms_list) == dtm.shape[1]
        assert vectorizer.terms_list == sorted(vectorizer.terms_list)

    def test_vectorizer_fixed_vocab(self, tokenized_docs):
        vocabulary_terms = ["lamb", "snow", "school", "rule", "teacher"]
        vectorizer = representations.Vectorizer(vocabulary_terms=vocabulary_terms)
        doc_term_matrix = vectorizer.fit_transform(tokenized_docs)
        assert len(vectorizer.vocabulary_terms) == len(vocabulary_terms)
        assert doc_term_matrix.shape[1] == len(vocabulary_terms)
        assert sorted(vectorizer.terms_list) == sorted(vocabulary_terms)

    def test_bad_transform(self, tokenized_docs):
        vectorizer = representations.Vectorizer()
        with pytest.raises(ValueError):
            _ = vectorizer.transform(tokenized_docs)


class TestGroupVectorizer:

    @pytest.mark.parametrize(
        "kwargs",
        [
            {"tf_type": "linear"},
            {"tf_type": "linear", "idf_type": "standard"},
            {"tf_type": "linear", "idf_type": "standard", "dl_type": "linear"},
            {"tf_type": "linear", "idf_type": "standard", "dl_type": "linear", "norm": "l1"},
            {"tf_type": "linear", "idf_type": "standard", "dl_type": "linear", "norm": "l2"},
            {"tf_type": "sqrt", "idf_type": "smooth", "dl_type": "sqrt"},
            {"tf_type": "log", "idf_type": "bm25", "dl_type": "log"},
            {"tf_type": "binary", "idf_type": None, "dl_type": None},
        ],
    )
    def test_weighting_combos(self, kwargs, tokenized_docs, groups):
        vectorizer = representations.GroupVectorizer(**kwargs)
        _ = vectorizer.fit(tokenized_docs, groups)
        _ = vectorizer.weighting

    def test_id_to_grp(self, grp_vectorizer_and_gtm):
        grp_vectorizer, _ = grp_vectorizer_and_gtm
        assert isinstance(grp_vectorizer.id_to_grp, dict)
        assert all(isinstance(key, int) for key in grp_vectorizer.id_to_grp.keys())
        assert all(
            isinstance(val, str) for val in grp_vectorizer.id_to_grp.values()
        )
        assert len(grp_vectorizer.id_to_grp) == len(grp_vectorizer.vocabulary_grps)

    def test_terms_and_grp_list(self, grp_vectorizer_and_gtm):
        grp_vectorizer, gtm = grp_vectorizer_and_gtm
        assert isinstance(grp_vectorizer.terms_list, list)
        assert isinstance(grp_vectorizer.grps_list, list)
        assert isinstance(grp_vectorizer.terms_list[0], str)
        assert len(grp_vectorizer.terms_list) == len(grp_vectorizer.vocabulary_terms)
        assert len(grp_vectorizer.terms_list) == gtm.shape[1]
        assert grp_vectorizer.terms_list == sorted(grp_vectorizer.terms_list)
        assert len(grp_vectorizer.grps_list) == len(grp_vectorizer.vocabulary_grps)
        assert len(grp_vectorizer.grps_list) == gtm.shape[0]
        assert grp_vectorizer.grps_list == sorted(grp_vectorizer.grps_list)

    def test_alt_fits(self, grp_vectorizer_and_gtm_2):
        grp_vectorizer, _ = grp_vectorizer_and_gtm_2
        # TODO: is this all we need???

    def test_fixed_vocab(self, tokenized_docs, groups):
        vocabulary_terms = ["lamb", "snow", "school", "rule", "teacher"]
        vocabulary_grps = ["a", "b"]
        grp_vectorizer = representations.GroupVectorizer(
            vocabulary_terms=vocabulary_terms, vocabulary_grps=vocabulary_grps
        )
        grp_term_matrix = grp_vectorizer.fit_transform(tokenized_docs, groups)
        assert len(grp_vectorizer.vocabulary_terms) == len(vocabulary_terms)
        assert grp_term_matrix.shape[1] == len(vocabulary_terms)
        assert sorted(grp_vectorizer.terms_list) == sorted(vocabulary_terms)
        assert len(grp_vectorizer.vocabulary_grps) == len(vocabulary_grps)
        assert grp_term_matrix.shape[0] == len(vocabulary_grps)
        assert sorted(grp_vectorizer.grps_list) == sorted(vocabulary_grps)

    def test_grp_vectorizer_bad_transform(self, tokenized_docs, groups):
        grp_vectorizer = representations.GroupVectorizer()
        with pytest.raises(ValueError):
            _ = grp_vectorizer.transform(tokenized_docs, groups)


def test_get_term_freqs(vectorizer_and_dtm, lamb_and_children_idxs):
    _, doc_term_matrix = vectorizer_and_dtm
    idx_lamb, idx_children = lamb_and_children_idxs
    term_freqs = representations.get_term_freqs(doc_term_matrix, type_="linear")
    assert len(term_freqs) == doc_term_matrix.shape[1]
    assert term_freqs.min() == 1
    assert term_freqs.max() == 5
    assert term_freqs[idx_lamb] == 5
    assert term_freqs[idx_children] == 2


def test_get_term_freqs_sublinear(vectorizer_and_dtm, lamb_and_children_idxs):
    _, doc_term_matrix = vectorizer_and_dtm
    idx_lamb, idx_children = lamb_and_children_idxs
    tfs = representations.get_term_freqs(doc_term_matrix, type_="linear")
    tfs_sqrt = representations.get_term_freqs(doc_term_matrix, type_="sqrt")
    tfs_log = representations.get_term_freqs(doc_term_matrix, type_="log")
    assert len(tfs) == len(tfs_sqrt) == len(tfs_log) == doc_term_matrix.shape[1]
    assert tfs_log.max() == pytest.approx(2.60943, abs=1e-3)
    assert tfs_log.min() == pytest.approx(1.0, abs=1e-3)
    assert tfs_log[idx_lamb] == pytest.approx(2.60943, abs=1e-3)
    assert tfs_log[idx_children] == pytest.approx(1.69314, abs=1e-3)
    assert (tfs_sqrt == np.sqrt(tfs)).all()
    assert (tfs_log == np.log(tfs) + 1.0).all()


def test_get_term_freqs_exception():
    with pytest.raises(ValueError):
        _ = representations.get_term_freqs(coo_matrix((1, 1)).tocsr())


def test_get_doc_freqs(vectorizer_and_dtm, lamb_and_children_idxs):
    _, doc_term_matrix = vectorizer_and_dtm
    idx_lamb, idx_children = lamb_and_children_idxs
    doc_freqs = representations.get_doc_freqs(doc_term_matrix)
    assert len(doc_freqs) == doc_term_matrix.shape[1]
    assert doc_freqs.max() == 5
    assert doc_freqs.min() == 1
    assert doc_freqs[idx_lamb] == 5
    assert doc_freqs[idx_children] == 2


def test_get_doc_freqs_exception():
    with pytest.raises(ValueError):
        _ = representations.get_doc_freqs(coo_matrix((1, 1)).tocsr())


def test_get_doc_lengths(tokenized_docs, vectorizer_and_dtm):
    _, doc_term_matrix = vectorizer_and_dtm
    dls = representations.get_doc_lengths(doc_term_matrix, type_="linear")
    assert len(dls) == doc_term_matrix.shape[0]
    for dl, td in zip(dls, tokenized_docs):
        assert dl == len(td)


def test_get_doc_lengths_type(vectorizer_and_dtm):
    _, doc_term_matrix = vectorizer_and_dtm
    dls = representations.get_doc_lengths(doc_term_matrix, type_="linear")
    dls_sqrt = representations.get_doc_lengths(doc_term_matrix, type_="sqrt")
    dls_log = representations.get_doc_lengths(doc_term_matrix, type_="log")
    assert len(dls) == len(dls_sqrt) == len(dls_log) == doc_term_matrix.shape[0]
    assert (dls_sqrt == np.sqrt(dls)).all()
    assert (dls_log == np.log(dls) + 1.0).all()


def test_get_doc_lengths_exception(vectorizer_and_dtm):
    _, doc_term_matrix = vectorizer_and_dtm
    with pytest.raises(ValueError):
        _ = representations.get_doc_lengths(doc_term_matrix, type_="foo")


def test_get_information_content(vectorizer_and_dtm, lamb_and_children_idxs):
    _, doc_term_matrix = vectorizer_and_dtm
    idx_lamb, idx_children = lamb_and_children_idxs
    ics = representations.get_information_content(doc_term_matrix)
    assert len(ics) == doc_term_matrix.shape[1]
    assert ((0.0 <= ics) & (ics <= 1.0)).all()
    assert ics.max() > ics.min()
    assert ics[idx_lamb] == pytest.approx(0.95443, rel=0.001)
    assert ics[idx_children] == pytest.approx(0.81127, rel=0.001)


def test_apply_idf_weighting(vectorizer_and_dtm):
    _, doc_term_matrix = vectorizer_and_dtm
    for type_ in ("standard", "smooth", "bm25"):
        reweighted_matrix = representations.apply_idf_weighting(doc_term_matrix, type_=type_)
        assert reweighted_matrix.shape == doc_term_matrix.shape


def test_filter_terms_by_df_identity(vectorizer_and_dtm):
    vectorizer, doc_term_matrix = vectorizer_and_dtm
    dtm, vocab = representations.filter_terms_by_df(
        doc_term_matrix,
        vectorizer.vocabulary_terms,
        max_df=1.0,
        min_df=1,
        max_n_terms=None,
    )
    assert dtm.shape == doc_term_matrix.shape
    assert vocab == vectorizer.vocabulary_terms


def test_filter_terms_by_df_max_n_terms(vectorizer_and_dtm):
    vectorizer, doc_term_matrix = vectorizer_and_dtm
    dtm, vocab = representations.filter_terms_by_df(
        doc_term_matrix,
        vectorizer.vocabulary_terms,
        max_df=1.0,
        min_df=1,
        max_n_terms=2,
    )
    assert dtm.shape[0] == doc_term_matrix.shape[0]
    assert dtm.shape[1] < doc_term_matrix.shape[1]
    assert all(term in vocab for term in ("lamb", "mary"))
    assert not any(term in vocab for term in ("children", "school", "teacher"))


def test_filter_terms_by_df_min_df(vectorizer_and_dtm):
    vectorizer, doc_term_matrix = vectorizer_and_dtm
    dtm, vocab = representations.filter_terms_by_df(
        doc_term_matrix,
        vectorizer.vocabulary_terms,
        max_df=1.0,
        min_df=2,
        max_n_terms=None,
    )
    assert dtm.shape[0] == doc_term_matrix.shape[0]
    assert dtm.shape[1] < doc_term_matrix.shape[1]
    assert all(term in vocab for term in ("children", "lamb", "mary", "school", "teacher"))


def test_filter_terms_by_df_exception(vectorizer_and_dtm):
    vectorizer, doc_term_matrix = vectorizer_and_dtm
    with pytest.raises(ValueError):
        _ = representations.filter_terms_by_df(
            doc_term_matrix,
            vectorizer.vocabulary_terms,
            max_df=1.0,
            min_df=6,
            max_n_terms=None,
        )



import re

import pytest
from spacy.tokens import Span

from textacy import load_spacy_lang
from textacy import extract
from textacy.extract.matches import _make_pattern_from_string


@pytest.fixture(scope="module")
def doc():
    nlp = load_spacy_lang("en_core_web_sm")
    text = (
        "Many years later, as he faced the firing squad, Colonel Aureliano Buendía was "
        "to remember that distant afternoon when his father took him to discover ice. "
        "At that time Macondo was a village of twenty adobe houses, built on the bank "
        "of a river of clear water that ran along a bed of polished stones, which were "
        "white and enormous, like prehistoric eggs. The world was so recent that many "
        "things lacked names, and in order to indicate them it was necessary to point."
    )
    return nlp(text)


class TestTokenMatches:

    @pytest.mark.parametrize(
        "patterns",
        [
            "POS:NOUN",
            ["POS:NOUN", "POS:DET"],
            [{"POS": "NOUN"}],
            [[{"POS": "NOUN"}], [{"POS": "DET"}]],
        ],
    )
    def test_pattern_types(self, doc, patterns):
        matches = list(extract.token_matches(doc, patterns))[:5]
        assert matches
        assert all(isinstance(span, Span) for span in matches)

    def test_patstr(self, doc):
        matches = list(extract.token_matches(doc, "POS:NOUN"))[:5]
        assert matches
        assert all(len(span) == 1 for span in matches)
        assert all(span[0].pos_ == "NOUN" for span in matches)

    def test_patstr_op(self, doc):
        matches = list(extract.token_matches(doc, "POS:NOUN:+"))[:5]
        assert matches
        assert all(len(span) >= 1 for span in matches)
        assert all(tok.pos_ == "NOUN" for span in matches for tok in span)

    def test_patstr_bool(self, doc):
        matches = list(extract.token_matches(doc, "IS_PUNCT:bool(True)"))[:5]
        assert matches
        assert all(span[0].is_punct is True for span in matches)

    def test_patstr_int(self, doc):
        matches = list(extract.token_matches(doc, "LENGTH:int(5)"))[:5]
        assert matches
        assert all(len(span[0]) == 5 for span in matches)

    def test_patdict(self, doc):
        matches = list(extract.token_matches(doc, [{"POS": "NOUN"}]))[:5]
        assert matches
        assert all(len(span) == 1 for span in matches)
        assert all(span[0].pos_ == "NOUN" for span in matches)

    def test_patdict_op(self, doc):
        matches = list(extract.token_matches(doc, [{"POS": "NOUN", "OP": "+"}]))[:5]
        assert matches
        assert all(len(span) >= 1 for span in matches)
        assert all(tok.pos_ == "NOUN" for span in matches for tok in span)

    def test_patdict_bool(self, doc):
        matches = list(extract.token_matches(doc, [{"IS_PUNCT": True}]))[:5]
        assert matches
        assert all(span[0].is_punct is True for span in matches)

    def test_patdict_int(self, doc):
        matches = list(extract.token_matches(doc, [{"LENGTH": 5}]))[:5]
        assert matches
        assert all(len(span[0]) == 5 for span in matches)

    @pytest.mark.parametrize(
        "patstr, pat",
        [
            ("TAG:VBZ", [{"TAG": "VBZ"}]),
            ("POS:NOUN:+", [{"POS": "NOUN", "OP": "+"}]),
            ("IS_PUNCT:bool(False)", [{"IS_PUNCT": False}]),
            (
                "IS_DIGIT:bool(True):? POS:NOUN:*",
                [{"IS_DIGIT": True, "OP": "?"}, {"POS": "NOUN", "OP": "*"}],
            ),
            (
                "LENGTH:int(5) DEP:nsubj:!",
                [{"LENGTH": 5}, {"DEP": "nsubj", "OP": "!"}],
            ),
            ("POS:DET :", [{"POS": "DET"}, {}]),
            (
                "IS_PUNCT:bool(False) : IS_PUNCT:bool(True)",
                [{"IS_PUNCT": False}, {}, {"IS_PUNCT": True}],
            ),
        ],
    )
    def test_make_pattern_from_string(self, patstr, pat):
        assert _make_pattern_from_string(patstr) == pat

    @pytest.mark.parametrize("patstr", ["POS", "POS:NOUN:VERB:+", "POS:NOUN:*?"])
    def test_make_pattern_from_str_error(self, patstr):
        with pytest.raises(ValueError):
            _ = _make_pattern_from_string(patstr)


class TestRegexMatches:

    @pytest.mark.parametrize(
        "pattern",
        [
            r"\w+ was",
            re.compile(r"\w+ was"),
            r"[Tt]he",
        ]
    )
    def test_pattern(self, pattern, doc):
        matches = list(extract.regex_matches(doc, pattern))
        assert matches
        assert all(isinstance(match, Span) for match in matches)

    @pytest.mark.parametrize("alignment_mode", ["strict", "contract", "expand"])
    def test_alignment_mode(self, alignment_mode, doc):
        pattern = r"\w [Tt]he \w"
        matches = list(
            extract.regex_matches(doc, pattern, alignment_mode=alignment_mode)
        )
        if alignment_mode == "strict":
            assert not matches
        elif alignment_mode == "contract":
            assert all(match.text.lower() == "the" for match in matches)
        elif alignment_mode == "expand":
            assert all("the" in match.text.lower() for match in matches)
            assert all(len(match.text) > 3 for match in matches)



"""
Hybrid Metrics
--------------

:mod:`textacy.similarity.hybrid`: Normalized similarity metrics that combine edit-,
token-, and/or sequence-based algorithms.
"""
from __future__ import annotations

from typing import Callable, Sequence

from . import edits
from .. import constants


def token_sort_ratio(s1: str | Sequence[str], s2: str | Sequence[str]) -> float:
    """
    Measure the similarity between two strings or sequences of strings
    using Levenshtein distance, only with non-alphanumeric characters removed
    and the ordering of tokens in each sorted before comparison.

    Args:
        s1
        s2

    Returns:
        Similarity between ``s1`` and ``s2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar strings.

    See Also:
        :func:`textacy.similarity.edits.levenshtein()`
    """
    str1 = _to_prepared_str(s1)
    str2 = _to_prepared_str(s2)
    return edits.levenshtein(str1, str2)


def _to_prepared_str(s: str | Sequence[str]) -> str:
    """
    Remove all characters from ``s`` except letters and numbers, strip whitespace,
    and force everything to lower-case; then sort tokens before re-joining into
    a single string.
    """
    tokens = (
        constants.RE_ALNUM.findall(s.lower())
        if isinstance(s, str)
        else [tok.lower().strip() for tok in s]
    )
    return " ".join(sorted(tokens))


def monge_elkan(
    seq1: Sequence[str],
    seq2: Sequence[str],
    sim_func: Callable[[str, str], float] = edits.levenshtein,
) -> float:
    """
    Measure the similarity between two sequences of strings using the (symmetric)
    Monge-Elkan method, which takes the average of the maximum pairwise similarity
    between the tokens in each sequence as compared to those in the other sequence.

    Args:
        seq1
        seq2
        sim_func: Callable that computes a string-to-string similarity metric;
            by default, Levenshtein edit distance.

    Returns:
        Similarity between ``seq1`` and ``seq2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar strings.

    See Also:
        :func:`textacy.similarity.edits.levenshtein()`
    """
    if not seq1 or not seq2:
        return 0.0

    sum_maxsim1 = sum(
        max(sim_func(tok1, tok2) for tok2 in seq2)
        for tok1 in seq1
    )
    sum_maxsim2 = sum(
        max(sim_func(tok2, tok1) for tok1 in seq1)
        for tok2 in seq2
    )
    return ((sum_maxsim1 / len(seq1)) + (sum_maxsim2 / len(seq2))) / 2
